# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
from __future__ import division, print_function
import unittest
import numpy as np
import math

from splearn.automaton import Automaton
from splearn.tests.datasets.get_dataset_path import get_dataset_path
from splearn.datasets.base import load_data_sample
from splearn import Spectral

class UnitaryTest(unittest.TestCase):

    def setUp(self):
        self.A = Automaton(2, 2, np.array([1.0, -1.0]),
                           np.array([0.0, 1.0]),
                           [np.array([[0.5, 1.0], [-0.5, 0.0]]),
                           np.array([[1.0, 1.0], [-2.0, 1/3]])])

    def test_constructor(self):
        # final assignment
        self.assertRaises(TypeError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          "final", [np.array([[0.5, 1.0], [-0.5, 0.0]]),
                                    np.array([[1.0, 1.0], [-2.0, 1/3]])])
        self.assertRaises(TypeError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0, 1]), [np.array([[0.5, 1.0],
                                                       [-0.5, 0.0]]),
                                             np.array([[1.0, 1.0],
                                                       [-2.0, 1/3]])])
        self.assertRaises(ValueError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0, 1.0]), [np.array([[0.5, 1.0],
                                                                [-0.5, 0.0]]),
                                                      np.array([[1.0, 1.0],
                                                                [-2.0, 1/3]])])
        # initial assignment
        self.assertRaises(TypeError, Automaton, 2, 2, "np.array([1.0, -1.0])",
                          np.array([0.0, 1.0]), [np.array([[0.5, 1.0],
                                                           [-0.5, 0.0]]),
                                                 np.array([[1.0, 1.0],
                                                           [-2.0, 1/3]])])
        self.assertRaises(TypeError, Automaton, 2, 2, np.array([1, -1]),
                          np.array([0.0, 1.0]), [np.array([[0.5, 1.0],
                                                           [-0.5, 0.0]]),
                                                 np.array([[1.0, 1.0],
                                                           [-2.0, 1/3]])])
        self.assertRaises(ValueError, Automaton, 2, 2,
                          np.array([1.0, -1.0, 0.0]),
                          np.array([0.0, 1.0]), [np.array([[0.5, 1.0],
                                                           [-0.5, 0.0]]),
                                                 np.array([[1.0, 1.0],
                                                           [-2.0, 1/3]])])
        # transitions assignment
        self.assertRaises(TypeError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), "transitions")
        self.assertRaises(ValueError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [1, 2, 3])
        self.assertRaises(TypeError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [1, 2])
        self.assertRaises(TypeError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.array([[0, 1],
                                                           [0, 0]]),
                                                 np.array([[1, 1],
                                                           [2, 1]])])
        self.assertRaises(ValueError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.ones((2, 2, 2),
                                                         dtype=np.float),
                                                 np.ones((2, 2, 2),
                                                         dtype=np.float)])
        self.assertRaises(ValueError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.ones((2, 3),
                                                         dtype=np.float),
                                                 np.ones((2, 3),
                                                         dtype=np.float)])
        self.assertRaises(ValueError, Automaton, 2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.ones((3, 3),
                                                         dtype=np.float),
                                                 np.ones((3, 3),
                                                         dtype=np.float)])
        # nbS assignment
        self.assertRaises(ValueError, Automaton, 2, 2.4, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.array([[0.5, 1.0],
                                                           [-0.5, 0.0]]),
                                                 np.array([[1.0, 1.0],
                                                           [-2.0, 1/3]])])
        self.assertRaises(ValueError, Automaton, 2, -2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.array([[0.5, 1.0],
                                                           [-0.5, 0.0]]),
                                                 np.array([[1.0, 1.0],
                                                           [-2.0, 1/3]])])
        # nbL assignment
        self.assertRaises(ValueError, Automaton, 2.4, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.array([[0.5, 1.0],
                                                           [-0.5, 0.0]]),
                                                 np.array([[1.0, 1.0],
                                                           [-2.0, 1/3]])])
        self.assertRaises(ValueError, Automaton, -2, 2, np.array([1.0, -1.0]),
                          np.array([0.0, 1.0]), [np.array([[0.5, 1.0],
                                                           [-0.5, 0.0]]),
                                                 np.array([[1.0, 1.0],
                                                           [-2.0, 1/3]])])

    def test_initial(self):
        np.testing.assert_array_equal(self.A.initial, np.array([1.0, -1.0]))

    def test_final(self):
        np.testing.assert_array_equal(self.A.final, np.array([0.0, 1.0]))

    def test_transitions(self):
        np.testing.assert_array_equal(self.A.transitions[0],
                                      np.array([[0.5, 1.0], [-0.5, 0.0]]))
        np.testing.assert_array_equal(self.A.transitions[1],
                                      np.array([[1.0, 1.0], [-2.0, 1/3]]))

    def test_nbS(self):
        self.assertEqual(self.A.nbS, 2)

    def test_nbL(self):
        self.assertEqual(self.A.nbL, 2)

    def test__rmul(self):
        B = 3.1 * self.A
        np.testing.assert_array_equal(B.initial, np.array([3.1, -3.1]))

    def test__add(self):
        # Test errors
        Ap = Automaton(3, 2, np.array([1.0, -1.0]),
                       np.array([0.0, 1.0]),
                       [np.array([[0.5, 1.0], [-0.5, 0.0]]),
                        np.array([[1.0, 1.0], [-2.0, 1/3]]),
                        np.array([[-1.0, 1.0], [-1.0, 1/6]])])
        self.assertRaises(ValueError, self.A.__add__, Ap)
        # Test addition validity
        B = self.A + self.A
        self.assertEqual(B.nbL, 2)
        self.assertEqual(B.nbS, 4)
        np.testing.assert_array_equal(B.initial,
                                      np.array([1.0, -1.0, 1.0, -1.0]))
        np.testing.assert_array_equal(B.final, np.array([0.0, 1.0, 0.0, 1.0]))
        np.testing.assert_array_equal(B.transitions[0],
                                      np.array([[0.5, 1.0, 0.0, 0.0],
                                                [-0.5, 0.0, 0.0, 0.0],
                                                [0.0, 0.0, 0.5, 1.0],
                                                [0.0, 0.0, -0.5, 0.0]]))

    def test_val(self):
        np.testing.assert_almost_equal(self.A.val([]), -1.0)
        np.testing.assert_almost_equal(self.A.val([0]), 1.0)
        np.testing.assert_almost_equal(self.A.val([1]), 2/3)
        np.testing.assert_almost_equal(self.A.val([0, 1]), 4/3)

    def test_mirror(self):
        B = self.A
        C = B.mirror()
        u = np.random.randint(0, B.nbL, 5)
        np.testing.assert_almost_equal(B.val(u), C.val(u[::-1]))

    def test_HouseholderReflector(self):
        v = np.array([3, 4])
        v = Automaton.HouseholderReflector(v)
        r = math.sqrt(5)/5*np.array([2, 1])
        np.testing.assert_array_almost_equal_nulp(v, r)
        v = np.array([-3, 4])
        v = Automaton.HouseholderReflector(v)
        r = math.sqrt(5) / 5*np.array([-2, 1])
        np.testing.assert_array_almost_equal_nulp(v, r)

    def test_mulHouseholderReflector(self):
        u = np.array([1, 2, 3])
        v = np.array([1, 2])
        w = Automaton.mulHouseholderReflector(u, v)
        np.testing.assert_array_almost_equal_nulp(w, np.array([1, -14, -29]))

    def test_HouseholderReductionFw(self):
        nbe = 2
        nbl = 2
        i = np.random.rand(nbe)
        t = np.random.rand(nbe)
        trans = [np.random.rand(nbe, nbe) for _ in range(nbl)]
        newNbe = 10
        p = np.random.rand(nbe, newNbe)
        ip = np.linalg.pinv(p)

        A = Automaton(nbl, newNbe, np.dot(i, p), np.dot(ip, t),
                      [np.dot(ip, np.dot(t, p)) for t in trans])
        B = A.HouseholderReductionFw(1.0e-5)
        # pr int(B.nbL,B.nbS,B.initial,B.final,B.transitions)
        self.assertEqual(B.nbS, nbe)
        np.testing.assert_almost_equal(A.val([]), B.val([]))
        np.testing.assert_almost_equal(A.val([0]), B.val([0]))
        np.testing.assert_almost_equal(A.val([0, 0]), B.val([0, 0]))
        np.testing.assert_almost_equal(A.val([1]), B.val([1]))
        np.testing.assert_almost_equal(A.val([0, 1, 0]), B.val([0, 1, 0]))
        np.testing.assert_almost_equal(A.val([0, 1, 0, 1, 1]),
                                       B.val([0, 1, 0, 1, 1]))

    def test_minimisation(self):
        nbe = 2
        nbl = 2
        i = np.random.rand(nbe)
        t = np.random.rand(nbe)
        trans = [np.random.rand(nbe, nbe) for _ in range(nbl)]
        newNbe = 10
        p = np.random.rand(nbe, newNbe)
        ip = np.linalg.pinv(p)

        A = Automaton(nbl, newNbe, np.dot(i, p), np.dot(ip, t),
                      [np.dot(ip, np.dot(t, p)) for t in trans])
        B = Automaton(nbl, nbe, np.zeros(nbe), np.random.rand(nbe),
                      [np.random.rand(nbe, nbe) for _ in range(nbl)])
        B = A+B
        C = Automaton(nbl, nbe, np.random.rand(nbe), np.zeros(nbe),
                      [np.random.rand(nbe, nbe) for _ in range(nbl)])
        B = B+C
        B = B.minimisation(1.0e-5)

        self.assertEqual(B.nbS, nbe)
        np.testing.assert_almost_equal(A.val([]), B.val([]))
        np.testing.assert_almost_equal(A.val([0]), B.val([0]))
        np.testing.assert_almost_equal(A.val([0, 0]), B.val([0, 0]))
        np.testing.assert_almost_equal(A.val([1]), B.val([1]))
        np.testing.assert_almost_equal(A.val([0, 1, 0]), B.val([0, 1, 0]))
        np.testing.assert_almost_equal(A.val([0, 1, 0, 1, 1]),
                                       B.val([0, 1, 0, 1, 1]))

    def test_absConv(self):
        nbe = 3
        nbl = 2
        i = np.random.rand(nbe)
        t = np.random.rand(nbe)
        trans = [np.random.rand(nbe, nbe)]
        trans.append(np.eye(nbe))
        A = Automaton(nbl, nbe, i, t, trans)
        self.assertEqual(A.isAbsConv, False)
        trans = [np.random.rand(nbe, nbe) for x in range(nbl)]
        i = i/i.sum()
        for q in range(nbe):
            s = t[q]
            for x in range(nbl):
                s += sum(trans[x][q, :])
            t[q] = t[q]/s
            for x in range(nbl):
                trans[x][q, :] = trans[x][q, :]/s
        A = Automaton(nbl, nbe, i, t, trans)
        self.assertEqual(A.isAbsConv, True)

    def test_sum(self):
        nbe = 3
        nbl = 2
        i = np.random.rand(nbe)
        t = np.random.rand(nbe)
        trans = [np.random.rand(nbe, nbe) for x in range(nbl)]
        i = i/i.sum()
        for q in range(nbe):
            s = t[q]
            for x in range(nbl):
                s += sum(trans[x][q, :])
            t[q] = t[q]/s
            for x in range(nbl):
                trans[x][q, :] = trans[x][q, :]/s
        A = Automaton(nbl, nbe, i, t, trans)
        np.testing.assert_almost_equal(A.sum(), 1)

    def test_load_Pautomac_Automaton(self):
        adr = get_dataset_path("pautomac3.txt")
        A = Automaton.load_Pautomac_Automaton(adr)
        r = A.nbL, A.nbS
        self.assertEqual(r, (4, 25))  # à compléter

    def test_SimpleExample(self):
        A = Automaton.SimpleExample()
        np.testing.assert_almost_equal(A.sum(), 1)

    def test_BuildHankels(self):
        A = Automaton.SimpleExample()
        lh = A.BuildHankels([(), (0,), (1,)],
                            [(), (0,), (1,)])
        self.assertEqual(lh[0][0, 0], 1/3)

    def test_transformation(self):
        A1 = Automaton.SimpleExample()
        t = A1.transformation(source='classic',target='prefix')
        A2 = t.transformation(source='prefix',target='classic')
        np.testing.assert_almost_equal(A1.final, A2.final)
        np.testing.assert_almost_equal(A1.initial, A2.initial)
        np.testing.assert_almost_equal(A1.transitions, A2.transitions)

    def test_calc_prefix_completion_weights(self):
        A = Automaton.SimpleExample()
        A = A.transformation()
        dic = A.calc_prefix_completion_weights([-1])
        self.assertEqual(set(dic.keys()), {0, 1, -1},
                         "test_calc_prefix_completion_weights failed")
    
    def test_get_dot(self):        
        train_file = '3.pautomac_light.train'
        data = load_data_sample(adr=get_dataset_path(train_file))
        sp = Spectral()
        sp.fit(X=data.data)
        dotfile = "3.pautomac_light.train.dot"
        dot = sp.automaton.get_dot(threshold = 0.2, title = dotfile)
        print(dot)
        gold_file = train_file + ".gv"
        gold_str = ""
        with open(get_dataset_path(gold_file), 'r') as f:
            gold_str = f.read()
        self.assertEqual(dot, gold_str, "test_get_dot failed")

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
"""This module contains the Serializer class
"""
import numpy as np

from splearn.automaton import Automaton
from splearn.hankel import Hankel
import scipy.sparse as sps

class Serializer(object):
    """ Serializer is an helping object for data serialization
    """

    @staticmethod
    def __serialize(data):
        if data is None or isinstance(data, (bool, int, float, str)):
            if type(data).__module__ == "numpy":
                return np.asscalar(data)
            return data
        if isinstance(data, list):
            return [Serializer.__serialize(val) for val in data]
        if isinstance(data, sps.dok_matrix):
            k_str = "({0:d},{1:d})"
            return {"scipy.dok_matrix": {"shape" : Serializer.__serialize(data.shape), "dtype":  str(data.dtype),
                                         "values" : Serializer.__serialize(dict(zip([k_str.format(i,j) for (i,j) in data.keys()],
                                                                                     data.values())))}}
        if isinstance(data, dict):
            if all(isinstance(k, str) for k in data):
                return {k: Serializer.__serialize(v) for k, v in data.items()}
            return {"dict": [[Serializer.__serialize(k), Serializer.__serialize(v)] for k, v in data.items()]}
        if isinstance(data, tuple):
            return {"tuple": [Serializer.__serialize(val) for val in data]}
        if isinstance(data, set):
            return {"set": [Serializer.__serialize(val) for val in data]}
        if isinstance(data, np.ndarray):
            return {"numpy.ndarray": {
                "values": data.tolist(),
                "dtype":  str(data.dtype)}}
        if isinstance(data, Automaton):
            data_dict = {"nbL":data.nbL, "nbS":data.nbS, "initial":data.initial, "final":data.final, 
                "transitions":data.transitions, "type":data.type}
            return {"automaton" : Serializer.__serialize(data_dict)}
        if isinstance(data, Hankel):
            data_dict = {"nbL":data.nbL, "lhankel" : data.lhankel, "version" : data.version,
                         "partial" : data.partial, "sparse" : data.sparse,
                         "build_from_sample" : data.build_from_sample, "nbEx" : data.nbEx}
            return {"hankel" : Serializer.__serialize(data_dict)}
        raise TypeError("Type %s is not serializabled" % type(data))

    @staticmethod
    def __restore_json(data_str):
        if "dict" in data_str:
            return dict(data_str["dict"])
        if "tuple" in data_str:
            return tuple(data_str["tuple"])
        if "set" in data_str:
            return set(data_str["set"])
        if "scipy.dok_matrix" in data_str:
            data = data_str["scipy.dok_matrix"]
            keys = {"shape", "dtype", "values"}
            if not keys.issubset(set(data.keys())):
                raise ValueError("The input data string (" + str(data_str) +
                                  ") should contain the following keys : \"" +
                                  '\", \"'.join(keys) + "\"")
            values = Serializer.__restore_json(data["values"])
            shape = Serializer.__restore_json(data["shape"])
            dok = sps.dok_matrix(shape, dtype=data["dtype"])
            for k, val in values.items():
                k = k.replace("(","").replace(")","")
                ind1, ind2 = k.split(",")
                dok[(int(ind1), int(ind2))] = val
            return dok
        if "numpy.ndarray" in data_str:
            data = data_str["numpy.ndarray"]
            keys = {"values", "dtype"}
            if not keys.issubset(set(data.keys())):
                raise ValueError("The input data string (" + str(data_str) +
                                  ") should contain the following keys : \"" +
                                  '\", \"'.join(keys) + "\"")
            return np.array(data["values"], dtype=data["dtype"])
        if "automaton" in data_str:
            data = Serializer.__restore_json(data_str["automaton"])
            keys = {"nbL", "nbS", "initial", "final", "transitions", "type"}
            if not keys.issubset(set(data.keys())):
                raise ValueError("The input data string (" + str(data_str) +
                                 ") should contain the following keys : \"" +
                                 '\", \"'.join(keys) + "\"")
            return Automaton(nbL=data["nbL"], nbS=data["nbS"], initial=data["initial"], final=data["final"],
                         transitions=data["transitions"], type=data["type"])
        if "hankel" in data_str:
            data = Serializer.__restore_json(data_str["hankel"])
            keys = {"nbL", "lhankel", "version", "partial", "sparse", "build_from_sample", "nbEx"}
            if not keys.issubset(set(data.keys())):
                raise ValueError("The input data string (" + str(data_str) +
                                 ") should contain the following keys : \"" +
                                 '\", \"'.join(keys) + "\"")
            H = Hankel(version=data["version"], partial=data["partial"], sparse=data["sparse"],
                       lhankel = data["lhankel"])
            if data["build_from_sample"]:
                H.build_from_sample = True
                H.nbL = data["nbL"]
                H.nbEx = data["nbEx"]
            return H
            
        return data_str
    
    @staticmethod
    def __restore_yaml(data_str):
        if data_str is None or isinstance(data_str, (bool, int, float, str)):
            return data_str
        if isinstance(data_str, list):
            return [Serializer.__restore_yaml(k) for k in data_str]
        if "dict" in data_str:
            return dict(data_str["dict"])
        if "tuple" in data_str:
            return tuple(data_str["tuple"])
        if "set" in data_str:
            return set(data_str["set"])
        if "scipy.dok_matrix" in data_str:
            data = data_str["scipy.dok_matrix"]
            keys = {"shape", "dtype", "values"}
            errorMsg = "The input data string (" + str(data_str)
            errorMsg += ") should contain the following keys : \"" + '\", \"'.join(keys) + "\""
            if data is None:
                raise ValueError(errorMsg)
            if not keys.issubset(set(data.keys())):
                raise ValueError(errorMsg)
            values = Serializer.__restore_json(data["values"])
            shape = Serializer.__restore_json(data["shape"])
            dok = sps.dok_matrix(shape, dtype=data["dtype"])
            for k, val in values.items():
                k = k.replace("(","").replace(")","")
                ind1, ind2 = k.split(",")
                dok[(int(ind1), int(ind2))] = val
            return dok
        if "numpy.ndarray" in data_str:
            data = data_str["numpy.ndarray"]
            keys = {"values", "dtype"}
            errorMsg = "The input data string (" + str(data_str)
            errorMsg += ") should contain the following keys : \"" + '\", \"'.join(keys) + "\""
            if data is None:
                raise ValueError(errorMsg)
            if not keys.issubset(set(data.keys())):
                raise ValueError(errorMsg)
            return np.array(data["values"], dtype=data["dtype"])
        if "automaton" in data_str:
            data = Serializer.__restore_yaml(data_str["automaton"])
            keys = {"nbL", "nbS", "initial", "final", "transitions", "type"}
            errorMsg = "The input data string (" + str(data_str)
            errorMsg += ") should contain the following keys : \"" + '\", \"'.join(keys) + "\""
            if data is None:
                raise ValueError(errorMsg)
            if not keys.issubset(set(data.keys())):
                raise ValueError(errorMsg)
            return Automaton(nbL=data["nbL"], nbS=data["nbS"], initial=Serializer.__restore_yaml(data["initial"]),
                             final=Serializer.__restore_yaml(data["final"]),
                             transitions=[Serializer.__restore_yaml(k) for k in data["transitions"]],
                             type=data["type"])
        if "hankel" in data_str:
            data = Serializer.__restore_yaml(data_str["hankel"])
            keys = {"nbL", "lhankel", "version", "partial", "sparse", "build_from_sample", "nbEx"}
            errorMsg = "The input data string (" + str(data_str)
            errorMsg += ") should contain the following keys : \"" + '\", \"'.join(keys) + "\""
            if data is None:
                raise ValueError(errorMsg)
            if not keys.issubset(set(data.keys())):
                raise ValueError(errorMsg)
            H = Hankel(version=data["version"], partial=data["partial"], sparse=data["sparse"],
                       lhankel = [Serializer.__restore_yaml(k) for k in data["lhankel"]])
            if data["build_from_sample"]:
                H.build_from_sample = True
                H.nbL = data["nbL"]
                H.nbEx = data["nbEx"]
            return H
        return data_str
    
    @staticmethod
    def data_to_json(data):
        """ return a string into json format that does contains the input data.
        
        - Input:
        
        :param data: data composed by any types that is serializabled
        
        - Output:
        
        :returns: the output string
        :rtype: str
        """
        
        import json
        return json.dumps(Serializer.__serialize(data))

    @staticmethod
    def json_to_data(json_data_str):
        """ return a data from input json string.
        
        - Input:
        
        :param json_data_str: the json input string
        
        - Output:
        
        :returns: the data
        :rtype: deduced form the json input string
        """
        
        import json
        return json.loads(json_data_str, object_hook=Serializer.__restore_json)
    
    @staticmethod
    def data_to_yaml(data):
        """ return a string into yaml format that does contains the input data.
        
        - Input:
        
        :param data: data composed by any types that is serializabled
        
        - Output:
        
        :returns: the output string
        :rtype: str
        """
        import yaml
        return yaml.dump(Serializer.__serialize(data))

    @staticmethod
    def yaml_to_data(yaml_data_str):
        """ return a data from input yaml string.
        
        - Input:
        
        :param yaml_data_str: the yaml input string
        
        - Output:
        
        :returns: the data
        :rtype: deduced form the yaml input string
        """
        import yaml
        return Serializer.__restore_yaml(yaml.load(yaml_data_str))



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########

from __future__ import division, print_function
import numpy as np
import unittest
from splearn.datasets.base import load_data_sample
from splearn.datasets.data_sample import DataSample, SplearnArray
from splearn.tests.datasets.get_dataset_path import get_dataset_path
from splearn.spectral import Spectral


class UnitaryTest(unittest.TestCase):

    def test_load_data_sample(self):
        adr = get_dataset_path("3.pautomac.train")
        # adr = get_dataset_path("essai")

        l = load_data_sample(adr=adr)
        nbL = l.nbL
        nbEx = l.nbEx
        sample = l.data
        self.assertEqual(nbL, 4)
        self.assertEqual(nbEx, 20000)
        nbEx2 = sample.shape[0]
        self.assertEqual(nbEx, nbEx2)
        lmax = sample.shape[1]
        self.assertEqual(67, lmax)


    def test_assignement(self):

        adr = get_dataset_path("essai")
        # adr = get_dataset_path("essai")

        s = load_data_sample(adr=adr)
        cl = Spectral()
        cl.polulate_dictionnaries(s.data)
        self.assertEqual(s.nbL,s.data.nbL)
        self.assertEqual(s.nbEx, s.data.nbEx)
        with self.assertRaises(TypeError):
            s.nbL= 2.0
        with self.assertRaises(ValueError):
            s.nbL = -1
        # nbEx assignment
        with self.assertRaises(TypeError):
            s.nbEx = 6.0

        with self.assertRaises(ValueError):
            s.nbEx = -6

    def test_load_Splearn_array(self):
        adr = get_dataset_path("3.pautomac.train")
        # adr = get_dataset_path("essai")

        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False)
        cl.polulate_dictionnaries(data.data)
        nbL = data.data.nbL
        nbEx = data.data.nbEx
        sample = data.data .sample
        pref = data.data.pref
        suff = data.data.suff

        self.assertEqual(nbL, 4)
        self.assertEqual(nbEx, 20000)
        nbEx2 = sum([sample[w] for w in sample])
        self.assertEqual(nbEx, nbEx2)
        nbPref1 = sum([sample[w]*(len(w)+1) for w in sample])
        nbPref2 = sum([pref[w] for w in pref])
        self.assertEqual(nbPref1, nbPref2)
        nbSuff1 = sum([sample[w]*(len(w)+1) for w in sample])
        nbSuff2 = sum([suff[w] for w in suff])
        self.assertEqual(nbSuff1, nbSuff2)

        cl = Spectral(version = 'factor', partial=False)
        cl.polulate_dictionnaries(data.data)
        fact = data.data.fact
        nbFact1 = sum([sample[w]*(len(w)+1)*(len(w)+2)/2 for w in sample])
        nbFact2 = sum([fact[w] for w in fact])
        self.assertEqual(nbFact1, nbFact2)

    def test_select_columns(self):
        adr = get_dataset_path("0.spice.train")
        pT = load_data_sample(adr=adr)
        cl = Spectral(partial=False)
        cl.polulate_dictionnaries(pT.data)
        # lR = pT.data.select_rows(nb_rows_max = 10, version = 'classic')
        # lC = pT.data.select_columns(nb_columns_max = 10, version = 'classic')
        # self.assertEqual(lR, [(), (3,), (3, 0), (3, 3), (3, 0, 3), (3, 1),
        #                       (3, 3, 0), (3, 0, 3, 0), (3, 1, 3),
        #                       (3, 3, 1)])
        # self.assertEqual(lC, [(), (3,), (0,), (3, 3), (3, 0), (0, 3), (2,),
        #                   (1,), (1, 3), (3, 0, 3)])

        cl = Spectral(version = 'prefix', partial=False)
        cl.polulate_dictionnaries(pT.data)
        # lRp = pT.data.select_rows(nb_rows_max = 10, version = 'prefix')
        # lCp = pT.data.select_columns(nb_columns_max = 10, version = 'prefix')
        # self.assertEqual(lRp, [(), (3,), (3, 0), (3, 0, 0), (3, 0, 1),
        #                        (3, 0, 2), (3, 0, 3), (3, 0, 0, 0),
        #                        (3, 0, 0, 1), (3, 0, 0, 2)])
        # self.assertEqual(lCp, [(), (3,), (0,), (1,), (3, 0), (3, 3), (2,),
        #                    (0, 3), (1, 3), (3, 1)])

        cl = Spectral(version = 'factor', partial=False)
        cl.polulate_dictionnaries(pT.data)
        # lRf = pT.data.select_rows(nb_rows_max = 10, version = 'factor')
        # lCf = pT.data.select_columns(nb_columns_max = 10, version = 'factor')
        # self.assertEqual(lRf,  [(), (3,), (0,), (1,), (3, 0), (3, 3), (2,),
        #                         (3, 1), (0, 3), (1, 3)])
        # self.assertEqual(lCf,  [(), (3,), (0,), (1,), (2,), (3, 0), (3, 3),
        #                         (1, 3), (0, 3), (3, 1)])


if __name__ == '__main__':
    unittest.main()



import pickle
import numpy as np
from splearn.datasets.data_sample import DataSample


def load_data_sample(adr, filetype='SPiCe', pickle=False):
    """Load a sample from file and returns a dictionary
    (word,count)

    - Input:

    :param str adr: address and name of the loaded file
    :param str filetype: (default value = 'SPiCe') indicate
           the structure of the file. Should be either 'SPiCe' or 'Pautomac'
    :param boolean pickle: if enabled it a pickle file is created from the loaded file. Default is fault.

    - Output:

    :returns: corresponding DataSample
    :rtype: DataSample


    :Example:

    >>> from splearn.datasets.base import load_data_sample
    >>> from splearn.tests.datasets.get_dataset_path import get_dataset_path
    >>> train_file = '3.pautomac_light.train' # '4.spice.train'
    >>> data = load_data_sample(adr=get_dataset_path(train_file))
    >>> data.nbL
    4
    >>> data.nbEx
    5000
    >>> data.data
    Splearn_array([[ 3.,  0.,  3., ..., -1., -1., -1.],
           [ 3.,  3., -1., ..., -1., -1., -1.],
           [ 3.,  2.,  0., ..., -1., -1., -1.],
           ...,
           [ 3.,  1.,  3., ..., -1., -1., -1.],
           [ 3.,  0.,  3., ..., -1., -1., -1.],
           [ 3.,  3.,  1., ..., -1., -1., -1.]])

    """

    if filetype == 'SPiCe' or filetype == 'Pautomac':
        data = _load_file_doublelecture(adr=adr, pickle=pickle)
        return DataSample(data=data)

def _load_file_doublelecture(adr, pickle=False):
    dsample = {}  # dictionary (word,count)
    _, max_length = _read_dimension(adr=adr)
    f = open(adr, "r")
    line = f.readline()
    l = line.split()
    nbEx = int(l[0])
    nbL = int(l[1])
    line = f.readline()
    data1 = np.zeros((nbEx, max_length ))
    data1 += -1
    i = 0
    while line:
        l = line.split()
        # w = () if int(l[0]) == 0 else tuple([int(x) for x in l[1:]])
        # dsample[w] = dsample[w] + 1 if w in dsample else 1
        # traitement du mot vide pour les préfixes, suffixes et facteurs
        w = [] if int(l[0]) == 0 else [int(x) for x in l[1:]]
        data1[i, :len(w)] = w
        line = f.readline()
        i += 1
    # print("data1 ", data1)
    f.close()
    if pickle:
        _create_pickle_files(adr=adr, dsample=dsample)
    return nbL, nbEx, data1

def _read_dimension(adr):
    f = open(adr, "r")
    line = f.readline()
    l = line.split()
    nbEx = int(l[0])
    nbL = int(l[1])
    line = f.readline()
    max_length = 0
    nb_sample = 0
    while line:
        l = line.split()
        nb_sample += 1
        length = int(l[0])
        if max_length < length:
            max_length = length
        line = f.readline()
    f.close()
    if nb_sample != nbEx:
        raise ValueError("check imput file, metadata " + str(nbEx) +
                         "do not match number of samples " + str(nb_sample))
    return nb_sample , max_length

def _create_pickle_files(self, adr, dsample):
    f = open(adr + ".sample.pkl", "wb")
    pickle.dump(dsample, f)
    f.close()



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########

from __future__ import division, print_function
import numpy as np
import unittest
from splearn.datasets.base import load_data_sample
from splearn.datasets.base import _read_dimension, _load_file_doublelecture
from splearn.tests.datasets.get_dataset_path import get_dataset_path


class UnitaryTest(unittest.TestCase):

    def test_load_data_sample(self):
        adr = get_dataset_path("3.pautomac.train")
        # adr = get_dataset_path("essai")

        l = load_data_sample(adr=adr)
        nbL = l.nbL
        nbEx = l.nbEx
        sample = l.data
        self.assertEqual(nbL, 4)
        self.assertEqual(nbEx, 20000)
        nbEx2 = sample.shape[0]
        self.assertEqual(nbEx, nbEx2)
        lmax = sample.shape[1]
        self.assertEqual(67, lmax)

    def test_read_dimension(self):
        adr = get_dataset_path("3.pautomac.train")
        # adr = get_dataset_path("essai")

        l = _read_dimension(adr=adr)
        nbEx = l[0]
        max_length = l[1]
        self.assertEqual(nbEx, 20000)
        self.assertEqual(max_length, 67)


    def test_load_file_doublelecture(self):
        adr = get_dataset_path("3.pautomac.train")
        l = _load_file_doublelecture(adr=adr)
        # adr = get_dataset_path("essai")
        nbL = l[0]
        nbEx = l[1]
        sample = l[2]
        self.assertEqual(nbL, 4)
        self.assertEqual(nbEx, 20000)
        nbEx2 = sample.shape[0]
        self.assertEqual(nbEx, nbEx2)
        lmax = sample.shape[1]
        self.assertEqual(67, lmax)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########

from __future__ import division, print_function
import numpy as np
import unittest
from splearn.datasets.base import load_data_sample
from splearn.automaton import Automaton
from splearn.spectral import Spectral
from splearn.tests.datasets.get_dataset_path import get_dataset_path
from sklearn.linear_model.tests.test_passive_aggressive import random_state


class SpectralTest(unittest.TestCase):

    def test_version(self):
        adr = get_dataset_path("essai")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, sparse=False, full_svd_calculation= True, version="prefix")
        X = data.data
        cl.fit(X=X)
        np.testing.assert_almost_equal(cl.automaton.initial,
                               np.array([-1.21159819e+00, -1.78488894e-01,
                                         1.30974719e-02, -1.66533454e-16,
                                         -3.70074342e-17]))

        cl = Spectral(partial=False, sparse=False, full_svd_calculation= True, version="suffix")
        cl.fit(X=X)
        np.testing.assert_almost_equal(cl.automaton.initial,
                                       np.array(
                                           [-3.16481225e-01, -1.95397796e-01,
                                            2.89623130e-01, 1.85499764e-15,
                                            0.00000000e+00]))
        cl = Spectral(partial=False, sparse=False,
                      version="factor", full_svd_calculation= True, mode_quiet=False)
        cl.fit(X=X)
        np.testing.assert_almost_equal(cl.automaton.final,
                                       np.array([ -4.35303631e-01,
                                                  1.49070953e+00,
                                                  4.80783716e-01,
                                                  3.75818126e-17,
                                                  0.00000000e+00]))

    def test_rank_fail(self):
        adr = get_dataset_path("essai")
        data = load_data_sample(adr=adr)
        X = data.data
        cl = Spectral(partial=True, lrows= 5, lcolumns=5,
                      sparse=False, version="prefix", rank=7)
        with self.assertRaises(ValueError):
            cl.fit(X=X)


    def test_non_automaton(self):
        cl = Spectral(partial=False, sparse=False)
        param = cl.get_params()
        self.assertEqual(param.get("partial"),False)
        X = np.array([[1, 2, 3, 6],[0, 0, 2, 5]])
        cl.fit(X=X)
        self.assertEqual(cl.automaton, None)

    def test_fit_spectral_notsparse(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, sparse=False)
        cl.fit(X=data.data)
        self.assertIsInstance(cl.automaton,Automaton,"fit function fails")

    def test_fit_spectral_sparse(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, sparse=True, rank=4)
        cl.fit(X=data.data)
        self.assertIsInstance(cl.automaton,Automaton,"fit function fails")


    def test_predict_spectral(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, sparse=False)
        cl.fit(X=data.data)
        XX = data.data[0:3,:]
        p = cl.predict(X=XX)
        np.testing.assert_almost_equal(
            p[0], np.array([ 0.16666667,  0.16666667,  0.16666667]))

    def test_loss_spectral(self):
        adr = get_dataset_path("essai")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, sparse=False, smooth_method = 'trigram')
        cl.fit(X=data.data)
        XX = data.data[0:3, :]
        p = cl.loss(X=XX, normalize=False)
        self.assertAlmostEquals(p, 5.375278407684168)
        p = cl.loss(X=XX, normalize=True)
        self.assertAlmostEquals(p, 1.7917594692280561)

    def test_score_spectral(self):
        adr = get_dataset_path("essai")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, sparse=False, smooth_method = 'trigram')
        cl.fit(X=data.data)
        XX = data.data[0:3, :]
        p = cl.score(X=XX, scoring="perplexity")
        self.assertAlmostEquals(p, -1.7917594692280561)
        p = cl.score(X=XX, scoring="none")
        self.assertAlmostEquals(p, -1.7917594692280561)


    def test_trigram(self):
        adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        # adr = get_dataset_path("essai")
        cl = Spectral(lrows=6, lcolumns=6, sparse=False,
                      partial=True)
        cl.set_params(smooth_method='trigram')
        cl.fit(data.data)
        self.assertEqual(cl.trigram[(1, 2)],
                         {0: 1543, 1: 1278, 2: 453, 3: 1254, -1: 5060, -2: 532} )
        with self.assertRaises(TypeError):
            cl.trigram = "dict"
        adr = get_dataset_path("essai")
        data = load_data_sample(adr=adr)
        cl = Spectral(lrows=6, lcolumns=6, sparse=False,
                      partial=True, smooth_method='trigram')
        cl.fit(data.data)
        cl.predict(data.data)
        np.testing.assert_equal(cl.trigram_index,
                                np.zeros(data.data.shape[0], dtype=bool))

        self.assertEqual(cl.nb_trigram(), 0)

    def test_trigramprobability(self):
        adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        # adr = get_dataset_path("essai")
        cl = Spectral(lrows=6, lcolumns=6, sparse=False,
                      partial=True)
        cl.set_params(smooth_method='trigram')
        cl.fit(data.data)
        lo = cl.loss(X=data.data)
        self.assertAlmostEqual(lo, 10.875367424508504)
        cl.nb_trigram()

    def test_loss_with_y(self):
        adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        # adr = get_dataset_path("essai")
        cl = Spectral(lrows=6, lcolumns=6, sparse=False,
                      partial=True)
        cl.set_params(smooth_method='trigram')
        cl.fit(data.data)
        p = cl.predict(X=data.data)
        lo = cl.loss(X=data.data, y=p, normalize=True)
        self.assertAlmostEqual(lo, 0.0)
        lo = cl.loss(X=data.data, y=p, normalize=False)
        self.assertAlmostEqual(lo, 0.0)


    def test_score_with_y(self):
        adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        # adr = get_dataset_path("essai")
        cl = Spectral(lrows=6, lcolumns=6, sparse=False,
                      partial=True)
        cl.set_params(smooth_method='trigram')
        cl.fit(data.data)
        p = cl.predict(X=data.data)
        lo = cl.score(X=data.data, y=p, scoring='perplexity')
        self.assertAlmostEqual(lo, 6220.871643894891)
        lo = cl.score(X=data.data, y=p, scoring='none')
        self.assertAlmostEqual(lo, -0.0)

    def test_smooth_method_none(self):
        adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        # adr = get_dataset_path("essai")
        cl = Spectral(lrows=6, lcolumns=6, sparse=False,
                      partial=True)
        cl.fit(data.data)
        p = cl.predict(X=data.data)
        with self.assertRaises(ValueError):
            cl.score(X=data.data)
        with self.assertRaises(ValueError):
            cl.score(X=data.data, y=p)
        self.assertAlmostEqual(cl.loss(X=data.data, y=p), 0.0)

    def test_smooth_method_notnone(self):
        with self.assertWarns(UserWarning):
            cl = Spectral(lrows=6, lcolumns=6, sparse=False,
                      partial=True, smooth_method='weighted')


    def test_BuildAutomatonFromHankel(self):  # a suivre
        A = Automaton.SimpleExample()
        H = A.to_hankel([(), (0,), (1,)], [(), (0,), (1,)], mode_quiet=False)
        B = H.to_automaton(rank=2, mode_quiet=False)
        np.testing.assert_almost_equal(A.val([]), B.val([]))
        np.testing.assert_almost_equal(A.val([0]), B.val([0]))
        np.testing.assert_almost_equal(A.val([0, 0]), B.val([0, 0]))
        np.testing.assert_almost_equal(A.val([1]), B.val([1]))
        np.testing.assert_almost_equal(A.val([0, 1, 0]), B.val([0, 1, 0]))
        np.testing.assert_almost_equal(A.val([0, 1, 0, 1, 1]),
                                       B.val([0, 1, 0, 1, 1]))

    def test_sklearn_compatibility(self):
        from sklearn.utils.estimator_checks import check_estimator
        from sklearn.model_selection import train_test_split, cross_val_score
        check_estimator(Spectral)
        adr = get_dataset_path("3.pautomac_light.train")
        data = load_data_sample(adr=adr)
        sp = Spectral(lrows=6, lcolumns=6, rank = 5, sparse=False,
                      partial=True, smooth_method='trigram')
        X_train, X_test = train_test_split(data.data, test_size=0.4, random_state=0)
        sp.fit(X_train)
        single_predicted_weights = sp.predict(X_test)
        print(single_predicted_weights)
        self.assertAlmostEqual(single_predicted_weights[0], 6.76217667e-02, delta = 1e-5)
        scores = cross_val_score(sp, data.data, cv=4)
        print(scores)
        scores_expected = [-10.65272755, -10.7090267,  -10.78404758, -11.08453211]
        for s1, s2 in zip(scores, scores_expected):
            self.assertAlmostEqual(s1, s2, delta=0.1)

#     def test_Perplexity(self):
#         adr = get_dataset_path("3.pautomac")
#         P = Learning()
#         P.load_Spice_Sample(adr + ".train")
#         A = P.LearnAutomaton(20, 7, 7, "factor", True)
#         P.Perplexity(A, adr)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
"""This module contains the DataSample class and SplearnArray class.


"""
import numpy as np


class SplearnArray(np.ndarray):
    """Sample data array used by the splearn spectral estimation

    **SplearnArray** class inherit from numpy ndarray as a 2d data ndarray.
    
    Example of a possible 2d shape:
    
    +---+---+---+---+---+
    |  0|  1|  0|  3| -1|
    +---+---+---+---+---+
    |  0|  0|  3|  3|  1|
    +---+---+---+---+---+
    |  1|  1| -1| -1| -1|
    +---+---+---+---+---+
    |  5| -1| -1| -1| -1|
    +---+---+---+---+---+
    | -1| -1| -1| -1| -1|
    +---+---+---+---+---+
    
    is equivalent to:
    
    - word (0103) or abad
    - word (00331) or aaddb
    - word (11) or bb
    - word (5) or f
    - word () or empty
    
    Each line represents a word of the sample. The words are represented by integer letters (0->a, 1->b, 2->c ...).
    -1 indicates the end of the word. The number of rows is the total number of words in the sample (=nbEx) and the number of columns
    is given by the size of the longest word. Notice that the total number of words does not care about the words' duplications. 
    If a word is duplicated in the sample, it is counted twice as two different examples. 
    
    The DataSample class encapsulates also the sample's parameters 'nbL', 'nbEx' (number of letters in the alphabet and 
    number of samples) and the fourth dictionaries 'sample', 'prefix', 'suffix' and 'factor' that will be populated during the fit
    estimations.
    
    - Input:

    :param nd.array input_array: input ndarray that will be converted into **SplearnArray**
    :param int nbL: the number of letters
    :param int nbEx: total number of examples.
    :param dict sample: the keys are the words and the values are the number of time it appears in the sample.
    :param dict pref: the keys are the prefixes and the values are the number of time it appears in the sample.
    :param dict suff: the keys are the suffixes and the values are the number of time it appears in the sample.
    :param dict fact: the keys are the factors and the values are the number of time it appears in the sample.

    :Example:

    >>> from splearn.datasets.base import load_data_sample
    >>> from splearn.tests.datasets.get_dataset_path import get_dataset_path
    >>> train_file = '3.pautomac_light.train' # '4.spice.train'
    >>> data = load_data_sample(adr=get_dataset_path(train_file))
    >>> print(data.__class__)
    >>> data.data
    <class 'splearn.datasets.data_sample.DataSample'>
    SplearnArray([[ 3.,  0.,  3., ..., -1., -1., -1.],
        [ 3.,  3., -1., ..., -1., -1., -1.],
        [ 3.,  2.,  0., ..., -1., -1., -1.],
        ...,
        [ 3.,  1.,  3., ..., -1., -1., -1.],
        [ 3.,  0.,  3., ..., -1., -1., -1.],
        [ 3.,  3.,  1., ..., -1., -1., -1.]])
    """
    def __new__(cls, input_array, nbL=None, nbEx=None,
                sample=None, pref=None,
                suff=None, fact=None, *args, **kwargs):
        obj = np.asarray(input_array).view(cls)
        obj.nbL = nbL
        obj.nbEx = nbEx
        obj.sample = sample
        obj.pref = pref
        obj.suff = suff
        obj.fact = fact
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.nbL = getattr(obj, 'nbL', None)
        self.nbEx = getattr(obj, 'nbEx', None)

        self.sample = getattr(obj, 'sample', None)
        self.pref = getattr(obj, 'pref', None)
        self.suff = getattr(obj, 'suff', None)
        self.fact = getattr(obj, 'fact', None)

class DataSample(dict):
    """ A DataSample instance

    - Input:

    :param tuple data: a tuple of (int, int, numpy.array) for the corresponding three elements
        (nbL, nbEx, data) where nbL is the number of letters in the alphabet, nbEx is the number
        of samples and data is the 2d data array

    :Example:

    >>> from splearn.datasets.base import load_data_sample
    >>> from splearn.tests.datasets.get_dataset_path import get_dataset_path
    >>> train_file = '3.pautomac_light.train' # '4.spice.train'
    >>> data = load_data_sample(adr=get_dataset_path(train_file))
    >>> print(data.__class__)
    <class 'splearn.datasets.data_sample.DataSample'>
    >>> data.nbL
    4
    >>> data.nbEx
    5000
    >>> data.data

    """

    def __init__(self, data=None, **kwargs):
        # The dictionary that contains the sample
        self._data = SplearnArray(np.zeros((0,0)))
        if data is not None:
            self.data = SplearnArray(data[2], nbL=data[0], nbEx=data[1])
        super(DataSample, self).__init__(kwargs)


    @property
    def nbL(self):
        """Number of letters"""
        return self.data.nbL

    @nbL.setter
    def nbL(self, nbL):
        if not isinstance(nbL, int):
            raise TypeError("nbL should be an integer")
        if nbL < 0:
            raise ValueError("The size of the alphabet should " +
                             "an integer >= 0")
        self.data.nbL = nbL

    @property
    def nbEx(self):
        """Number of examples"""
        return self.data.nbEx

    @nbEx.setter
    def nbEx(self, nbEx):
        if not isinstance(nbEx, int):
            raise TypeError("nbEx should be an integer")
        if nbEx < 0:
            raise ValueError("The number of examples should be " +
                             " an integer >= 0")
        self.data.nbEx = nbEx

    @property
    def data(self):
        """SplearnArray"""
        return self._data

    @data.setter
    def data(self, data):
        if isinstance(data, (SplearnArray, np.ndarray, np.generic)):
            self._data = data
        else:
            raise TypeError("sample should be a SplearnArray.")




# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
import unittest
import numpy as np
import os
from collections import deque
import yaml

from splearn.automaton import Automaton
from splearn.hankel import Hankel
from splearn.serializer import Serializer
from splearn.spectral import Spectral
from splearn.tests.datasets.get_dataset_path import get_dataset_path
from splearn.datasets.base import load_data_sample

class SerializerTest(unittest.TestCase):


    def setUp(self):
        self.A = Automaton.SimpleExample()
        self.title = "Simple Example"
        self.input_file = 'simple_example'
        self.formats = ['json', 'yaml']
        self.words = [[], [0], [0, 0], [1], [1, 1], [0, 1, 0], [1, 0, 1]]
        self.d_str = "Value for word \"{:s}\" = {:.2f}"
  
    def testReadAutomaton(self):
        for f in self.formats:
            B = Automaton.read(get_dataset_path(self.input_file + '.' + f), format=f)
            for w in self.words:
                np.testing.assert_almost_equal(self.A.val(w), B.val(w))
    
    def testWriteAutomata(self):
        for f in self.formats:
            Automaton.write(self.A, get_dataset_path(self.input_file + '_2.' + f), format=f)
            B = Automaton.read(get_dataset_path(self.input_file + '_2.' + f), format=f)
            for w in self.words:
                np.testing.assert_almost_equal(self.A.val(w), B.val(w))
        for f in self.formats:
            os.remove(get_dataset_path(self.input_file + '_2.' + f))
    
    def testReadHankel(self):
        for f in self.formats:
            H = self.A.to_hankel([(), (0,), (1,)], [(), (0,), (1,)])
            Hankel.write(H, get_dataset_path(self.input_file + "_hankel" + "." + f), format=f)
            Hb = Hankel.read(get_dataset_path(self.input_file + "_hankel" + "." + f), format = f)
            B = Hb.to_automaton(2)
            for w in self.words:
                np.testing.assert_almost_equal(self.A.val(w), B.val(w))
        for f in self.formats:
            os.remove(get_dataset_path(self.input_file + "_hankel" + "." + f))
    
    def testReadWriteRealHankel(self):
        adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        X = data.data
        sp = Spectral()
        sp = sp.fit(X)
        H = sp.hankel
        for f in self.formats:
            Hankel.write(H, get_dataset_path("3.pautomac.train" + "_hankel" + "." + f), format=f)
            Hb = Hankel.read(get_dataset_path("3.pautomac.train" + "_hankel" + "." + f), format = f)
            self.assertTrue(H == Hb)
        for f in self.formats:
            os.remove(get_dataset_path("3.pautomac.train" + "_hankel" + "." + f))
    
    def testOthersSerializationTypes(self):
        data = [{'a' : 10, 40 : 'gu'}, {'toto', 5, 2.5, 'b'}, ('gh', 25, 'ko', 1.0)]
        data_json_str = Serializer.data_to_json(data)
        data_yaml_str = Serializer.data_to_yaml(data)
        data_json = Serializer.json_to_data(data_json_str)
        data_yaml = Serializer.yaml_to_data(data_yaml_str)
        self.assertEqual(data, data_json)
        self.assertEqual(data, data_yaml)
        
        data = [1, 2, 3.0]
        data_json_str = Serializer.data_to_json(data)
        data_yaml_str = Serializer.data_to_yaml(data)
        data_json = Serializer.json_to_data(data_json_str)
        data_yaml = Serializer.yaml_to_data(data_yaml_str)
        self.assertEqual(data, data_json)
        self.assertEqual(data, data_yaml)
    
    def testBadTypeSerialieationException(self):
        with self.assertRaises(TypeError):
            Serializer.data_to_json(deque('ghi'))
    
    def testBadDataException(self):
        yamlstrList = ["- scipy.dok_matrix:", 
                       "- scipy.dok_matrix:\n    shape:\n        tuple: [1, 1]\n    values: {'(0,0)': 1.0}",
                       "- numpy.ndarray:",
                       "- numpy.ndarray:\n    dtype: float64",
                       "- automaton:",
                       "- automaton:\n    nbL: 1",
                       "- hankel:",
                       "- hankel:\n    nbL: 1"]
        jsonstrList = ["{\"scipy.dok_matrix\":{}}",
                       "{\"numpy.ndarray\":{}}",
                       "{\"automaton\":{}}",
                       "{\"hankel\":{}}"]
        for yamlstr in yamlstrList:
            with self.assertRaises(ValueError):
                Serializer.yaml_to_data(yamlstr)
        for jsonstr in jsonstrList:
            with self.assertRaises(ValueError):
                Serializer.json_to_data(jsonstr)

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
"""This module contains the Hankel class

"""
from __future__ import division, print_function
import scipy.sparse as sps
import scipy.sparse.linalg as lin
import numpy as np
from numpy.linalg import svd, pinv
from sklearn.utils.extmath import randomized_svd as sk_svd

class Hankel(object):
    """ A Hankel instance , compute the list of Hankel matrices

    - Input:
    
    :param SplearnArray sample_instance: instance of SplearnArray
    :param lrows: number or list of rows,
           a list of strings if partial=True;
           otherwise, based on self.pref if version="classic" or
           "prefix", self.fact otherwise
    :type lrows: int or list of int
    :param lcolumns: number or list of columns
           a list of strings if partial=True ;
           otherwise, based on self.suff if version="classic" or "suffix",
           self.fact otherwise
    :type lcolumns: int or list of int
    :param string version: (default = "classic") version name
    :param boolean partial: (default value = False) build of partial
    :param boolean sparse: (default value = False) True if Hankel
           matrix is sparse
    :param boolean full_svd_calculation: (default value = False) if True the entire SVD is calculated
           for building hankel matrix. Else it is done by the sklearn random algorithm only for the greatest
           k=rank eigenvalues.
    :param boolean mode_quiet: (default value = False) True for no
           output message.
    :param list lhankel: list of all Hankel matrices. At least one of the two parameters 
           *sample_instance* or *lhankel* has to be not None. If *sample_instance* is given,
           the **Hankel** instance is built directly from the sample dictionnary,
           else it is deduced from the *lhankels* list of matrices. 

    :Example:
    
    >>> from splearn import Learning, Hankel , Spectral
    >>> train_file = '0.spice.train'
    >>> pT = load_data_sample(adr=train_file)
    >>> sp = Spectral()
    >>> sp.fit(X=pT.data)
    >>> lhankel = Hankel( sample_instance=pT.sample,
    >>>                   nbL=pT.nbL, nbEx=pT.nbEx,
    >>>                   lrows=6, lcolumns=6, version="classic",
    >>>                   partial=True, sparse=True, mode_quiet=True).lhankel

    """

    def __init__(
            self, sample_instance=None,
            lrows=[], lcolumns=[],
            version="classic", partial=False,
            sparse=False, full_svd_calculation=False, mode_quiet=False, lhankel=None):
        
        self.version = version
        self.partial = partial
        self.sparse = sparse
        self.full_svd__calculation = full_svd_calculation
        self.build_from_sample = True
        if sample_instance is not None:
            # Size of the alphabet
            self.nbL = sample_instance.nbL
            # Number of samples
            self.nbEx = sample_instance.nbEx
            self.lhankel = self.build(sample=sample_instance.sample,
                                      pref=sample_instance.pref,
                                      suff=sample_instance.suff,
                                      fact=sample_instance.fact,
                                      lrows=lrows, lcolumns=lcolumns,
                                      mode_quiet=mode_quiet)
        elif lhankel is not None:
            # Size of the alphabet
            self.nbL = len(lhankel) - 1
            self.lhankel = lhankel
            self.build_from_sample = False
            self._nbEx = -1
        else:
            raise ValueError("At least sample_instance or lhankel has to be not None.")

    def __eq__(self, other):
        #print("Hankel equality check")
        if self.version != other.version:
        #    print("version is different")
            return False
        if self.partial != other.partial:
        #    print("partial is different")
            return False
        if self.sparse != other.sparse:
        #    print("sparse is different")
            return False
        if self.build_from_sample != other.build_from_sample:
        #    print("build_from_sample is different")
            return False
        if self.nbL != other.nbL:
        #    print("nbL is different")
            return False
        if self.nbEx != other.nbEx:
        #    print("nbEx is different")
            return False
        if len(self.lhankel) != len(other.lhankel):
        #    print("lhankel length is different")
            return False
        for lh1, lh2 in zip(self.lhankel, other.lhankel):
            if self.sparse:
                if (lh1 != lh2).nnz > 0:
                #    print("{:d} elements oh lhandel are different".format((lh1 != lh2).nnz))
                    return False
            elif not np.array_equal(lh1, lh2):
            #    print("Different Array")
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def nbL(self):
        """Number of letters"""
        return self._nbL

    @nbL.setter
    def nbL(self, nbL):
        if not isinstance(nbL, int):
            raise TypeError("nbL should be an integer")
        if nbL < 0:
            raise ValueError("The size of the alphabet should " +
                             "an integer >= 0")
        self._nbL = nbL

    @property
    def nbEx(self):
        """Number of examples"""
        return self._nbEx

    @nbEx.setter
    def nbEx(self, nbEx):
        if not isinstance(nbEx, int):
            raise TypeError("nbEx should be an integer")
        if nbEx < 0:
            raise ValueError("The number of examples should be " +
                             " an integer >= 0")
        self._nbEx = nbEx
    
    @property
    def build_from_sample(self):
        """Boolean that indicates if the matrices have been build form sample or not
        (directly build from an Automaton in this case) """
        return self._build_from_sample
    
    @build_from_sample.setter
    def build_from_sample(self, val):
        if val:
            self._build_from_sample = True
        else:
            self._build_from_sample = False
        

    def build(self, sample, pref, suff, fact, lrows, lcolumns, mode_quiet):

        """ Create a Hankel matrix

        - Input:

        :param dict sample: the keys are the words and the values are the number of time it appears in the sample.
        :param dict pref: the keys are the prefixes and the values are the number of time it appears in the sample.
        :param dict suff: the keys are the suffixes and the values are the number of time it appears in the sample.
        :param dict fact: the keys are the factors and the values are the number of time it appears in the sample.
        :param lrows: number or list of rows,
               a list of strings if partial=True;
               otherwise, based on self.pref if version="classic" or
               "prefix", self.fact otherwise
        :type lrows: int or list of int
        :param lcolumns: number or list of columns
               a list of strings if partial=True ;
               otherwise, based on self.suff if version="classic" or "suffix",
               self.fact otherwise
        :type lcolumns: int or list of int
        :param boolean mode_quiet:  True for no output message.

        - Output:

        :returns: list lhankel, list of hankel matrix,
                  a DoK based sparse matrix or nuppy matrix based not sparse
        :rtype: list of matrix

        """
        # calcul des lignes lrows et colonnes lcolumns
        if not mode_quiet:
            print("Start Hankel matrix computation")
        if not self.partial:
            (lrows, lcolumns) = self._build_not_partial(
                pref=pref, suff=suff, fact=fact)
        else:
            (lrows, lcolumns) = self._build_partial(
                pref=pref, suff=suff, fact=fact,
                lrows=lrows, lcolumns=lcolumns)

        lhankel = self._create_hankel(sample=sample, pref=pref,
                                      suff=suff, fact=fact,
                                      lrows=lrows, lcolumns=lcolumns)
        if not mode_quiet:
            print ("End of Hankel matrix computation")
        return lhankel

    def _build_not_partial(self,pref, suff, fact):
        version = self.version
        if version == "classic":
            lrows = pref.keys()
            lcolumns = suff.keys()
        elif version == "prefix":
            lrows = pref.keys()
            lcolumns = fact.keys()
        elif version == "suffix":
            lrows = fact.keys()
            lcolumns = suff.keys()
        else:
            lrows = fact.keys()
            lcolumns = fact.keys()

        return (lrows, lcolumns)

    def _build_partial(self,
                       pref, suff, fact,
                       lrows, lcolumns):
        version = self.version

        if version == "classic":
            (lrows, lcolumns) = self._construc_partial_lrows_lcolumns(
                dict_first=pref, dict_second=suff,
                lrows=lrows, lcolumns=lcolumns)
        elif version == "prefix":
            (lrows, lcolumns) = self._construc_partial_lrows_lcolumns(
                dict_first=pref, dict_second=fact,
                lrows=lrows, lcolumns=lcolumns)
        elif version == "suffix":
            (lrows, lcolumns) = self._construc_partial_lrows_lcolumns(
                dict_first=fact, dict_second=suff,
                lrows=lrows, lcolumns=lcolumns)
        else:
            (lrows, lcolumns) = self._construc_partial_lrows_lcolumns(
                dict_first=fact, dict_second=fact,
                lrows=lrows, lcolumns=lcolumns)
        return lrows, lcolumns

    def _construc_partial_lrows_lcolumns(self, dict_first, dict_second,
                                         lrows,
                                         lcolumns):

        if isinstance(lrows, int):
            longmax = lrows
            lrows = [w for w in dict_first if len(w) <= longmax]
        else:
            s_first = set(dict_first)  # corresponding set
            lrows = [w for w in lrows if w in s_first]
        if isinstance(lcolumns, int):
            longmax = lcolumns
            lcolumns = [w for w in dict_second if len(w) <= longmax]
        else:
            s_second = set(dict_second)  # corresponding set
            lcolumns = [w for w in lcolumns if w in s_second]
        return (lrows, lcolumns)

    def _create_hankel(self, sample, pref, suff, fact, lrows, lcolumns):
        version = self.version
        sparse = self.sparse

        (drows, dcolumns) = self._sorted_rows_columns(lrows, lcolumns)

        nbRows = len(lrows)
        nbColumns = len(lcolumns)
        srows = set(lrows)
        scolumns = set(lcolumns)

        if sparse:
            lhankel = [sps.dok_matrix((nbRows, nbColumns)) for
                       i in range(self.nbL+1)]
        else:
            lhankel = [np.zeros((nbRows, nbColumns)) for
                       k in range(self.nbL+1)]
        if version == "classic":
            dsample = sample
        elif version == "prefix":
            dsample = pref
        elif version == "suffix":
            dsample = suff
        else:
            dsample = fact
        for w in dsample:
            for i in range(len(w)+1):
                if w[:i] in srows:
                    if w[i:] in scolumns:
                        lhankel[0][drows[w[:i]], dcolumns[w[i:]]] = dsample[w]
                    if (i < len(w) and w[i+1:] in scolumns):
                        lhankel[w[i]+1][drows[w[:i]],
                                        dcolumns[w[i+1:]]] = dsample[w]
        return lhankel

    def _sorted_rows_columns(self, lrows, lcolumns):
        nbRows = len(lrows)
        nbColumns = len(lcolumns)
        lrows = sorted(lrows, key=lambda x: (len(x), x))
        drows = {lrows[i]: i for i in range(nbRows)}
        lcolumns = sorted(lcolumns, key=lambda x: (len(x), x))
        dcolumns = {lcolumns[i]: i for i in range(nbColumns)}

        return (drows, dcolumns)
    
    def to_automaton(self, rank, mode_quiet=False):
        """ Return an automaton from the current Hankel matrix

        - Input:

        :param int rank: the matrix rank
        :param boolean mode_quiet: True for no output message.

        - Output:

        :returns: An automaton instance
        :rtype: Automaton
        """
        from splearn.automaton import Automaton
        if not mode_quiet:
            print("Start Building Automaton from Hankel matrix")
        matrix_shape =min(self.lhankel[0].shape)
        if (min(self.lhankel[0].shape) < rank) :
            raise ValueError("The value of parameter rank ("+  str(rank)
                             + ") should be less than " +
                              "the smaller dimension of the Hankel Matrix (" +
                             str(matrix_shape) + ")")
        if not self.sparse:
            hankel = self.lhankel[0]
            if self.full_svd__calculation:
                [u, s, v] = svd(hankel)
                u = u[:, :rank]
                v = v[:rank, :]
                # ds = np.zeros((rank, rank), dtype=complex)
                ds = np.diag(s[:rank])
            else:
                [u, s, v] = sk_svd(hankel, n_components=rank)
                ds = np.diag(s)
            pis = pinv(v)
            del v
            pip = pinv(np.dot(u, ds))
            del u, ds
            init = np.dot(hankel[0, :], pis)
            term = np.dot(pip, hankel[:, 0])
            trans = []
            for x in range(self.nbL):
                hankel = self.lhankel[x+1]
                trans.append(np.dot(pip, np.dot(hankel, pis)))

        else:
            hankel = self.lhankel[0].tocsr()
            if self.full_svd__calculation:
                [u, s, v] = svd(hankel.A)
                u = u[:, :rank]
                v = v[:rank, :]
                # ds = np.zeros((rank, rank), dtype=complex)
                ds = np.diag(s[:rank])
            else:
                [u, s, v] = lin.svds(hankel, k=rank)
                ds = np.diag(s)
            pis = pinv(v)
            del v
            pip = pinv(np.dot(u, ds))
            del u, ds
            init = hankel[0, :].dot(pis)[0, :]
            term = np.dot(pip, hankel[:, 0].toarray())[:, 0]
            trans = []
            for x in range(self.nbL):
                hankel = self.lhankel[x+1].tocsr()
                trans.append(np.dot(pip, hankel.dot(pis)))

        A = Automaton(nbL=self.nbL, nbS=rank, initial=init, final=term,
                         transitions=trans, type=self.version)
        if self.build_from_sample:
            A.initial = A.initial / self.nbEx
            if self.version == "prefix":
                A = A.transformation(source="prefix", target="classic")
            if self.version == "factor":
                A = A.transformation(source="factor", target="classic")
            if self.version == "suffix":
                A = A.transformation(source="suffix", target="classic")
            if not mode_quiet:
                print ("End of Automaton computation")
        return A

    @staticmethod
    def write(hankel_in, filename, format='json'):
        """ write input hankel into a file with the given format.

        - Input:

        :param Hankel hankel_in: hankel to write into the file
        :param str filename: the name of the file. If it does not exist,
         the file is created.
        :param str format: 'json' or yaml'
        """
        from splearn.serializer import Serializer
        if format == 'json':
            data_str = Serializer.data_to_json(hankel_in)
        elif format == 'yaml':
            data_str = Serializer.data_to_yaml(hankel_in)
        else:
            raise ValueError("Invalid input format. Should be \"json\" or \"yaml\"")
        with open(filename, 'w') as outfile:
            outfile.write(data_str)

    @staticmethod
    def read(filename, format='json'):
        """ return a Hankel build with attributes read from a file

        - Input:

        :param str filename: the name of the input file.
        :param str format: 'json' or yaml'

        - Output:

        :returns: the output hankel
        :rtype: Hankel
        """
        from splearn.serializer import Serializer
        with open(filename) as infile:
            datastr = infile.read()
        if format == 'json':
            return Serializer.json_to_data(datastr)
        if format == 'yaml':
            return Serializer.yaml_to_data(datastr)
        raise ValueError("Invalid input format. Should be \"json\" or \"yaml\"")



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
"""This module contains the Automaton class
"""

import numpy as np


class Automaton(object):
    """ Define an automaton with parameters

    - Input:

    :param int nbL: the number of letters
    :param int nbS: the number of states
    :param list initial: the initial vector
    :param list final: the final vector
    :param list transition: the transitions tables


    """

    def __init__(self, nbL=0, nbS=0, initial=[], final=[],
                 transitions=[], type='classic'):
        # The number of letters
        self.nbL = nbL
        # The number of states
        self.nbS = nbS
        # The vector containing the initial weight of each state
        self.initial = initial
        # The vector containing the final weight of each state
        self.final = final
        # The list of arrays defining the transitions
        self.transitions = transitions
        # The type of automaton
        self.type = type
        # Flag if the absolute convergence of the automaton has been calculated
        # or not
        self.__convcalculated = False
        self.__isAbsConv = False

    @property
    def final(self):
        """The vector containing the final weight of each state"""
        return self.__final

    @final.setter
    def final(self, final_values):
        if (not isinstance(final_values, np.ndarray) or
           final_values.dtype.type != np.float_):
            mess = "final_values should be a numpy.array of float.\n"
            mess += "Actual : " + str(final_values)
            raise TypeError(mess)
        if len(final_values) != self.nbS:
            raise ValueError("final_values length should be " + str(self.nbS))
        self.__final = final_values
        self.__convcalculated = False

    @property
    def initial(self):
        """The vector containing the initial weight of each state"""
        return self.__initial

    @initial.setter
    def initial(self, initial_values):
        if (not isinstance(initial_values, np.ndarray) or
           initial_values.dtype.type != np.float_):
            raise TypeError("initial_values should be a numpy.array of float" +
                            str(initial_values))
        if len(initial_values) != self.nbS:
            raise ValueError("initial_values length should be " +
                             str(self.nbS))
        self.__initial = initial_values
        self.__convcalculated = False

    @property
    def transitions(self):
        """The list of arrays defining the transitions"""
        return self.__transitions

    @transitions.setter
    def transitions(self, transitions_values):
        if not isinstance(transitions_values, list):
            raise TypeError("transitions_values should be a list")
        if len(transitions_values) != self.nbL:
            raise ValueError("The length of transitions_values should be " +
                             str(self.nbL))
        for x in transitions_values:
            if not isinstance(x, np.ndarray) or x.dtype != np.float_:
                raise TypeError("transitions_values should contain" +
                                " numpy.array of float")
            if (len(x.shape) != 2 or x.shape[0] != x.shape[1] or
                    x.shape[0] != self.nbS):
                mess = "Elements of transitions_value should "
                mess += "be {0:d}x{0:d} square matrices."
                mess = mess.format(self.nbS)
                raise ValueError(mess)
        self.__transitions = transitions_values
        self.__convcalculated = False

    @property
    def nbS(self):
        """The number of states"""
        return self.__nbS

    @nbS.setter
    def nbS(self, nbS_value):
        if (nbS_value == int(nbS_value) and nbS_value > 0):
            self.__nbS = int(nbS_value)
            self.__convcalculated = False
        else:
            raise ValueError("Error in new nbS value.")

    @property
    def nbL(self):
        """The number of letters"""
        return self.__nbL

    @nbL.setter
    def nbL(self, nbL_value):
        if (nbL_value == int(nbL_value) and nbL_value > 0):
            self.__nbL = int(nbL_value)
            self.__convcalculated = False
        else:
            raise ValueError("Error in new nbL value.")

    @property
    def isAbsConv(self):
        """Does the automaton meet the sufficient condition to be absolutely convergent"""
        if not self.__convcalculated:
            self._calcAbsConv()
            self.__convcalculated = True
        return self.__isAbsConv

    @isAbsConv.setter
    def isAbsConv(self, isAbsConvL_value):
        pass


    @property
    def type(self):
        """The string indicates the type of automaton"""
        return self._type

    @type.setter
    def type(self, type_value):
        if (not isinstance(type_value, str)):
            mess = "type_value should be a string.\n"
            mess += "Actual : " + str(type_value)
            raise TypeError(mess)
        if (type_value == 'classic' or  type_value == 'prefix' or
                    type_value == 'suffix' or type_value == 'factor'):
            self._type = type_value
        else:
            raise ValueError("type must be classic, prefix, suffix or factor.")

    def __rmul__(self, coeff):
        """ Multiplication of an automaton by a scalar

        - Input:

        :param scalar coeff: the coefficient of multiplication

        - Output:

        :returns: An automaton instance
        :rtype: Automaton

        """
        initial = coeff*self.initial
        return(Automaton(self.nbL, self.nbS, initial, self.final,
                         self.transitions))

    def __add__(self, automaton):
        """ Computes self + automaton
        self and automaton are built on the same alphabet

        - Input:

        :param Automaton automaton: automaton to add

        - Output:

        :returns: An automaton instance
        :rtype: Automaton

        """
        if self.nbL != automaton.nbL:
            raise ValueError("Can't add Automaton with different alphabets")
        else:
            initial = np.concatenate((self.initial, automaton.initial))
            final = np.concatenate((self.final, automaton.final))
            nbS = self.nbS + automaton.nbS
            nbL = self.nbL
            transitions = [np.zeros((nbS, nbS)) for _ in range(nbL)]
            for i in range(len(transitions)):
                transitions[i][0:self.nbS, 0:self.nbS] =\
                    self.transitions[i][0:self.nbS, 0:self.nbS]
                transitions[i][self.nbS:nbS, self.nbS:nbS] =\
                    automaton.transitions[i][0:automaton.nbS, 0:automaton.nbS]

        return Automaton(nbL, nbS, initial, final, transitions)

    @staticmethod
    def SimpleExample():
        """ A Probabilistic Automaton with two states and two letters.

        - Output:

        :returns: An automaton instance example with simple values
        :rtype: Automaton
        """
        initial = np.array([0.5, 0.5])
        final = np.array([1/2, 1/6])
        transitions = [np.array([[1/6, 1/12], [1/6, 1/6]]),
                       np.array([[0, 1/4], [1/6, 1/3]])]
        return Automaton(2, 2, initial, final, transitions)

    def transformation(self, source="classic", target="prefix"):
        """ Takes an automaton as input and transforms it.

        - Input:

        :param str source: "prefix", "factor" or "classic" or "suffix"(default)
        :param str target: "prefix" (default) "factor" or "classic" or "suffix"

        - Output:

        :returns: The result automaton instance
        :rtype: Automaton

         The transformation is done according to the source and target
         parameters.
         .. warning:: it does not check the convergence
         """
        A = Automaton(nbL=self.nbL, nbS=self.nbS, initial=self.initial,
                      final=self.final,
                      transitions=self.transitions)
        m_sigma = np.zeros(self.nbS)
        for m in self.transitions:
            m_sigma = m_sigma + m
        m = np.eye(self.nbS) - m_sigma
        im = np.linalg.inv(m)
        if source == "classic":
            if target == "prefix" or target == "factor":
                A.final = np.dot(im, A.final)
            if target == "factor" or target == "suffix":
                A.initial = np.dot(A.initial, im)
        elif target == "classic" :
            if source == "prefix" or source == "factor":
                A.final = np.dot(m, A.final)
            if source == "factor" or source == "suffix":
                A.initial = np.dot(A.initial, m)
        A.type = target
        return A

    def BuildHankels(self, lrows=[], lcolumns=[]):
        """ Return all Hankel (denses) matrices built on lrows and lcolumns from an automaton

        - Input:

        :param list lrows:
        :param list lcolumns:

        - Output:

        :returns: list of all Hankel matrices built on lrows and lcolumns
        :rtype: list
        """
        nbrows = len(lrows)
        nbcolumns = len(lcolumns)
        lh = [np.zeros((nbrows, nbcolumns)) for _ in range(self.nbL + 1)]
        dval = {}  # dictionary of already calculated values
        for i in range(nbrows):
            for j in range(nbcolumns):
                w = lrows[i] + lcolumns[j]
                if w in dval:
                    lh[0][i, j] = dval[w]
                else:
                    lh[0][i, j] = self.val(w)
                    dval[w] = lh[0][i, j]
                for x in range(self.nbL):
                    w = lrows[i] + (x,) + lcolumns[j]
                    if w in dval:
                        lh[x+1][i, j] = dval[w]
                    else:
                        lh[x+1][i, j] = self.val(w)
                        dval[w] = lh[x+1][i, j]
        return lh
    
    def to_hankel(self, lrows, lcolumns, mode_quiet=False):
        """ Return an Hankel instance (denses, classic and not partial) with matrices
        built on lrows and lcolumns from an automaton

        - Input:

        :param list lrows: prefixes
        :param list lcolumns: suffixes
        :param boolean mode_quiet: (default value = False) True for no
               output message.

        - Output:

        :returns: Hankel instance
        :rtype: Hankel
        """
        from splearn.hankel import Hankel
        lhankel = self.BuildHankels(lrows, lcolumns)
        return Hankel(mode_quiet=mode_quiet, lhankel=lhankel)

    def mirror(self):
        """ Compute the mirror automaton

        - Input:

        :param Automaton self: Automaton(nbL, nbS, initial, final, transitions)

        - Output:

        :returns: mA = Automaton(nbL, nbS, final, initial, Newtransitions)
                where Newtransitions[x] = transpose(transitions[x])
        :rtype: Automaton

        """
        Newtransitions = []
        for x in range(self.nbL):
            Newtransitions.append(np.transpose(self.transitions[x]))
        return Automaton(self.nbL, self.nbS, self.final, self.initial,
                         Newtransitions)

    def val(self, word):
        """ Compute the value computed by the automaton on word

        - Input:

        :param Automaton self: weighted automaton
        :param str word: a string

        - Output:

        :returns: probability r_A(w)
        :rtype: float

        """
        u = self.initial
        final = self.final
        for x in word:
            u = np.dot(u, self.transitions[x])
        return np.dot(u, final)

    @staticmethod
    def HouseholderReflector(x):
        """ the vector which defines the Householder for x

        - Input:

        :param vector x: a vector in :math:`R^k` different from 0

        - Output:

        :returns: :math:`v = u/||u||` 
            where :math:`u_1 = x_1 + sign(x_1)||x||`
            and :math:`u_i = x_i` for :math:`i \\geq 2`
        :rtype: vector

        """

        u = x.copy()
        s = 1 if x[0] >= 0 else -1
        u[0] = x[0] + s * np.linalg.norm(x)
        return u/np.linalg.norm(u)

    @staticmethod
    def mulHouseholderReflector(u, v):
        """ the product of u by the HouseholderReflector nxn matrix based on v

        - Input:

        :param vector u: row vector of :math:`R^n`
        :param vector v: vector of :math:`R^k` (k<=n)

        - Output:

        :returns: w, row vector of :math:`R^n`,
             :math:`w = uP(v)` where
             :math:`P(v)=[I_{n-k} 0; 0 R]\in R^{n \\times n}` and
             :math:`R=I_k-2v^T.v`
        :rtype: vector

        """
        n = len(u)
        w = u.copy()
        k = len(v)
        r = np.dot(u[n-k:n], v)
        w[n-k:n] = w[n-k:n] - 2*r*v
        return w

    def HouseholderReductionFw(self, tau):
        """ algorithm (Fig. 3) from the paper
        Stability and complexity of Minimising Probabilistic Automata
        by Kiefer and Wachter

        - Input:

        :param Automaton self: an object of the automaton class
        :param float tau: error tolerance parameter >=0

        - Output:

        :returns: The canonical forward reduction computed to the tolerance tau
        :rtype: Automaton

        """

        alpha = self.initial  # initial vector
        n = len(alpha)  # initial dimension
        v = Automaton.HouseholderReflector(alpha/np.linalg.norm(alpha))
        # vP1\in R^n
        lv = [v]  # list of Householder reflectors
        l = 0
        j = 1  # new number of dimensions
        e = np.zeros(n)
        e[0] = 1
        f = Automaton.mulHouseholderReflector(e, v)  # f \in R^n
        F = np.zeros([n, n])  # the projection matrix
        F[j-1, :] = f
        transNew = [np.zeros([n, n]) for x in range(self.nbL)]
        # new transitions
        while l < j:
            l += 1
            for x in range(self.nbL):
                f = F[l-1, :]
                u = np.dot(f, self.transitions[x])  # f_lM(a) u\in R^n
                for v in lv:
                    u = Automaton.mulHouseholderReflector(u, v)  # u \in R^n
                transNew[x][l-1, :] = u
                if (j+1 <= n and np.linalg.norm(transNew[x][l-1][j:n]) > tau):
                    j += 1
                    v = Automaton.HouseholderReflector(transNew[x][l-1][j-1:n])
                    # v \in R^{n-j}
                    lv.append(v)
                    transNew[x][l-1, :] = Automaton.mulHouseholderReflector(
                                          transNew[x][l-1, :], v)
                    f = np.zeros(n)
                    f[j-1] = 1
                    for v in reversed(lv):
                        f = Automaton.mulHouseholderReflector(f, v)
                    F[j-1, :] = f
        F = F[:j, :]
        for x in range(self.nbL):
            transNew[x] = transNew[x][:j, :j]
        iF = np.linalg.pinv(F)
        alphaNew = np.dot(alpha, iF)
        etaNew = np.dot(F, self.final)

        return Automaton(self.nbL, j, alphaNew, etaNew, transNew)

    def minimisation(self, tau):
        """ compute an equivalent minimal automaton, to the precision tau

        - Input:

        :param Automaton self:

        - Output:

        :returns: B, equivalent to A with a minimal number of states
        :rtype: Automaton

        """

        A = self.HouseholderReductionFw(tau)
        A = A.mirror()
        A = A.HouseholderReductionFw(tau)
        return A.mirror()

    def _calcAbsConv(self):
        """ a sufficient condition to be absolutely convergent

        - Input:

        :param Automaton self:

        - Output:

        :returns: False if  :math:`\\Sigma r_|A|(w)` is not convergent and
                True otherwise.
                It may happen that A is abs conv and that the return value
                is False
        :rtype: bool

        """
        m = np.zeros([self.nbS, self.nbS])
        for x in range(self.nbL):
            m = m + abs(self.transitions[x])
        if max(abs(np.linalg.eigvals(m))) < 1:
            self.__isAbsConv = True
        else:
            self.__isAbsConv = False

    def sum(self):
        """ the sum of a rational series

        - Input:

        :param Automaton self:

        - Output:

        :returns: sum over all samples of transitions
        :rtype: ndarray
        """
        m = np.zeros([self.nbS, self.nbS])
        for x in range(self.nbL):
            m = m + self.transitions[x]
        return np.dot(np.dot(self.initial, np.linalg.inv(np.eye(self.nbS)-m)),
                      self.final)

    @staticmethod
    def load_Pautomac_Automaton(adr):
        """ Load an automaton from a PAutomaC file and returns an object of the
        class Automaton; works for PFA and PDFA - not for HMM.

        - Input:

        :param string adr: address and name of the loaden file

        - Output

        :returns: An automaton instance
        :rtype: Automaton
        
        """
        states = set()  # set of states
        letters = set()  # alphabet
        dinit = {}  # dictionary of (initial state, initial value)
        dfinal = {}  # dictionary of (final state, final value)
        trans1 = {}  # [(state,letter), total weight] ; sum_l w(q,l) = 1
        trans2 = {}  # [(state,letter,state) weight] ; sum_q' w(q,l,q')=1
        f = open(adr, "r")
        f.readline()
        line = f.readline()
        # load dinit
        while line and line.find("F") == -1:
            line = line.replace('(', ' ').replace(')', ' ')
            l = line.split()
            q = int(l[0])
            states.add(q)
            dinit[q] = float(l[1])
            line = f.readline()
        line = f.readline()
        # load dfinal
        while line and line.find("S") == -1:
            line = line.replace('(', ' ').replace(')', ' ').replace(',', ' ')
            l = line.split()
            q = int(l[0])
            states.add(q)
            dfinal[q] = float(l[1])
            line = f.readline()
        line = f.readline()
        # load trans1
        while line and line.find("T") == -1:
            line = line.replace('(', ' ').replace(')', ' ').replace(',', ' ')
            l = line.split()
            q = int(l[0])
            states.add(q)
            x = int(l[1])
            letters.add(x)
            trans1[(q, x)] = float(l[2])
            line = f.readline()
        line = f.readline()
        # load trans2
        while line and line.find("(") != -1:
            line = line.replace('(', ' ').replace(')', ' ').replace(',', ' ')
            l = line.split()
            q1 = int(l[0])
            states.add(q1)
            x = int(l[1])
            letters.add(x)
            q2 = int(l[2])
            states.add(q2)
            trans2[(q1, x, q2)] = float(l[3])
            line = f.readline()
        f.close()
        nbl = max(list(letters)) + 1  # size of the alphabet
        nbs = max(list(states)) + 1  # number of states
        init = np.zeros(nbs)  # initial states
        for q in dinit.keys():
            init[q] = dinit[q]
        final = np.zeros(nbs)  # final states
        for q in dfinal.keys():
            final[q] = dfinal[q]
        trans = [np.zeros([nbs, nbs]) for x in range(nbl)]
        for (q1, x, q2) in trans2:
            if (q1, x) in trans1.keys():
                trans[x][q1, q2] = trans2[q1, x, q2] * trans1[q1, x] *\
                                   (1-final[q1])
        A = Automaton(nbl, nbs, init, final, trans)
        return A

    def calc_prefix_completion_weights(self, prefix):
        """ For the SPiCe competition for instance

        - Input:

        :param Automaton self: Be careful that A should be a prefix transformation of an Automata.
                               (see :func:`~automaton.Automaton.transformation`)
        :param List prefix: list of integers representing a prefix

        - Output:

        :returns: a dictionary with all alphabet letters as keys. The
                  associated values are the weights of being the next letter.
        :rtype: dict

        """
        if not self.isAbsConv:
            raise ValueError("The automaton absolute convergence is not true.")
        # Symbol -1 corresponds to the end of the sequence
        # If the weight is negative it does not carry any semantic
        p_w = self.val(prefix)
        toReturn = {}
        for i in range(self.nbL):
            w = self.val(prefix+[i])
            p_w -= w
            toReturn.update({i: max(w, 0)})
        toReturn.update({-1: max(p_w, 0)})
        return toReturn
    
    def get_dot(self, threshold = 0., nb_dec = 2, title = "Weighted Automata"):
        """ Return a string that contains the Automata into dot (graphviz) format
        
        :Example:

        >>> from splearn.datasets.base import load_data_sample
        >>> from splearn.tests.datasets.get_dataset_path import get_dataset_path
        >>> from splearn import Spectral
        >>> train_file = '3.pautomac_light.train'
        >>> data = load_data_sample(adr=get_dataset_path(train_file))
        >>> sp = Spectral()
        >>> sp.fit(X=data.data)
        >>> dotfile = "3.pautomac_light.train.dot"
        >>> dot = sp.Automaton.get_dot(threshold = 0.2, title = dotfile)
        >>> # To display the dot string one can use graphviz:
        >>> from graphviz import Source
        >>> src = Source(dot)
        >>> src.render(dotfile + '.gv', view=True) 

        - Input:

        :param  Automaton self
        :param float threshold for the value to keep. If \|weight\| < threshold, the 
        corresponding transition is not kept as an edge in the final dot string.
        :param int nb_dec is the number of decimals to keep for the weights.
        :param string title corresponds to the top comment of the string
        
        :returns: a string with the current Automata in dot format 
        """
        prec = ".{:d}f".format(nb_dec)
        out = "//{:s}\ndigraph {{\n".format(title)
        for i in range(self.nbS):
            if np.abs(self.initial[i]) > threshold and np.abs(self.final[i]) > threshold:
                label = "{0:d}\n______\n> {1:" + prec + "}\n{2:" + prec + "} >"
                label = label.format(i, self.initial[i], self.final[i])
                out += "\t{0:d} [label=\"".format(i)
                out += label + "\"]\n"
            elif np.abs(self.initial[i]) > threshold:
                label = "{0:d}\n______\n> {1:" + prec + "}"
                label = label.format(i, self.initial[i])
                out += "\t{0:d} [label=\"".format(i)
                out += label + "\"]\n"
            elif np.abs(self.final[i]) > threshold:
                label = "{0:d}\n______\n{1:" + prec + "} >"
                label = label.format(i, self.final[i])
                out += "\t{0:d} [label=\"".format(i)
                out += label + "\"]\n"
            else:                
                label = "{0:d}\n______"
                label = label.format(i)
                out += "\t{0:d} [label=\"".format(i)
                out += label + "\"]\n"
        for l in range(self.nbL):
            for i in range(self.nbS):
                for j in range(self.nbS):
                    weight = self.transitions[l][i,j]
                    if np.abs(weight) > threshold:
                        label = "{0:d}:{1:" + prec + "}"
                        label = label.format(l,weight)
                        out += "\t{0:d} -> {1:d} [label=\"".format(i,j)
                        out += label + "\"]\n"
        out += "}\n"
        return out

    @staticmethod
    def write(automaton_in, filename, format='json'):
        """ write input automaton into a file with the given format.

        - Input:

        :param Automaton automaton_in: automaton to write into the file
        :param str filename: the name of the file. If it does not exist,
         the file is created.
        :param str format: 'json' or yaml'
        """
        from splearn.serializer import Serializer
        if format == 'json':
            data_str = Serializer.data_to_json(automaton_in)
        elif format == 'yaml':
            data_str = Serializer.data_to_yaml(automaton_in)
        else:
            raise ValueError("Invalid input format. Should be \"json\" or \"yaml\"")
        with open(filename, 'w') as outfile:
            outfile.write(data_str)

    @staticmethod
    def read(filename, format='json'):
        """ return an Automaton build with attributes read from a file

        - Input:

        :param str filename: the name of the input file.
        :param str format: 'json' or yaml'

        - Output:

        :returns: the output automaton
        :rtype: Automaton
        """
        from splearn.serializer import Serializer
        with open(filename) as infile:
            datastr = infile.read()
        if format == 'json':
            return Serializer.json_to_data(datastr)
        if format == 'yaml':
            return Serializer.yaml_to_data(datastr)
        raise ValueError("Invalid input format. Should be \"json\" or \"yaml\"")
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########

from __future__ import print_function
import os
import sys
import shutil

# Always prefer setuptools over distutils
try:
    from setuptools import setup
    from setuptools.command.clean import clean
    from setuptools.command.sdist import sdist
except ImportError:
    from distutils.core import setup
    from distutils.command.clean import clean
    from distutils.command.sdist import sdist

USE_COPYRIGHT = True
try:
    from copyright import writeStamp, eraseStamp
except ImportError:
    USE_COPYRIGHT = False


###################
# Get scikit-splearn version
####################
def get_version():
    v_text = open('VERSION').read().strip()
    v_text_formted = '{"' + v_text.replace('\n', '","').replace(':', '":"')
    v_text_formted += '"}'
    v_dict = eval(v_text_formted)
    return v_dict["splearn"]


########################
# Set ltfat __version__
########################
def set_version(splearn_dir, version):
    filename = os.path.join(splearn_dir, '__init__.py')
    buf = ""
    for line in open(filename, "rb"):
        if not line.decode("utf8").startswith("__version__ ="):
            buf += line.decode("utf8")
    f = open(filename, "wb")
    f.write(buf.encode("utf8"))
    f.write(('__version__ = "%s"\n' % version).encode("utf8"))


######################
# Custom clean command
######################
class m_clean(clean):
    """ Remove build directories, and compiled file in the source tree"""

    def run(self):
        clean.run(self)
        if os.path.exists('build'):
            shutil.rmtree('build')
        for dirpath, dirnames, filenames in os.walk('splearn'):
            for filename in filenames:
                if (filename.endswith('.so') or
                        filename.endswith('.pyd') or
                        filename.endswith('.dll') or
                        filename.endswith('.pyc')):
                    os.unlink(os.path.join(dirpath, filename))
            for dirname in dirnames:
                if dirname == '__pycache__':
                    shutil.rmtree(os.path.join(dirpath, dirname))


##############################
# Custom sdist command
##############################
class m_sdist(sdist):
    """ Build source package

    WARNING : The stamping must be done on an default utf8 machine !
    """

    def run(self):
        if USE_COPYRIGHT:
            writeStamp()
            sdist.run(self)
            # eraseStamp()
        else:
            sdist.run(self)


##########################
# File path read command
##########################
def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r', encoding='utf-8') as f:
        return f.read()


####################
# Setup method
####################
def setup_package():
    """ Setup function"""
    # set version
    VERSION = get_version()
    splearn_dir = 'splearn'
    set_version(splearn_dir, VERSION)

    setup(name="scikit-splearn",
          version=VERSION,
          description='Python module for spectral learning of weighted automata',
          long_description=(read('README.rst') + '\n\n' +
                            read('HISTORY.rst') + '\n\n' +
                            read('AUTHORS.rst')),
          packages=["splearn", "splearn.datasets", "splearn.tests", "splearn.tests.datasets"],
          package_data={'splearn.tests.datasets': ['*']},
          url="https://gitlab.lis-lab.fr/dev/scikit-splearn.git",
          license='new BSD',
          author='François Denis and Rémi Eyraud and Denis Arrivault and Dominique Benielli',
          author_email='francois.denis@lis-lab.fr, ' +
                       'remi.eyraud@lis-lab.fr, ' +
                       'denis.arrivault@lis-lab.fr, ' +
                       'dominique.benielli@univ-amu.fr ',
          cmdclass={'clean': m_clean, 'sdist': m_sdist},
          classifiers=['Development Status :: 5 - Production/Stable',
                       'Intended Audience :: Science/Research',
                       'Intended Audience :: Developers',
                       'Natural Language :: English',
                       'License :: OSI Approved :: BSD License',
                       'Operating System :: MacOS :: MacOS X',
                       'Operating System :: POSIX :: Linux',
                       'Programming Language :: Python :: 3.4',
                       'Topic :: Scientific/Engineering',
                       'Topic :: Scientific/Engineering :: Mathematics'
                       ],
          install_requires=['numpy>=1.8', 'scipy>=0.16', 'six>=1.10', 'scikit-learn>=0.17.1'],
          )

if __name__ == "__main__":
    setup_package()



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
"""Module to get the absolute path of a reference dataset for tests

.. moduleauthor:: Denis Arrivault
"""

from __future__ import print_function, division

import os


def get_dataset_path(filename):
    """Return the absolute path of a reference dataset for tests

    - Input parameter:

    :param str filename: File name of the file containing reference data
        for tests (which must be in ``skgilearn/tests/datasets/``)

    - Output parameters:

    :returns: The absolute path where the file with name **filename** is stored
    :rtype: str

    """

    datasets_path = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(datasets_path, filename)



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########
"""This module contains the Spectral and Learning class

.. module author:: François Denis

"""
from __future__ import division, print_function
import numpy as np
import math
import warnings
from concurrent.futures import ThreadPoolExecutor, wait
import threading
lock = threading.RLock()


from splearn.datasets.data_sample import SplearnArray
from splearn.hankel import Hankel
from sklearn.base import BaseEstimator
from sklearn.utils import check_array
from sklearn.utils.validation import NotFittedError


class Spectral(BaseEstimator):
    """A Spectral estimator instance

    - Input:

    :param int rank: the ranking number
    :param lrows: (default value = 7) number or list of rows
           a list of strings or an interger indicating the max length
           of elements to consider if partial=True
           otherwise, based on self.pref if version="classic" or
           "prefix", self.fact otherwise
    :type lrows: int or tuple of int
    :param lcolumns: (default value = 7) number or list of columns
           a list of strings or an interger indicating the max length
           of elements to consider if partial=True
           otherwise, based on self.suff if version="classic" or "suffix",
           self.fact otherwise
    :type lcolumns: int or tuple of int
    :param string version: (default value = "classic") version name
    :param boolean partial: (default value = False) build
           of partial Hankel matrix
    :param boolean sparse: (default value = False) True if Hankel
           matrix is sparse
    :param boolean full_svd_calculation: (default value = False) if True the entire SVD is calculated
           for building hankel matrix. Else it is done by the sklearn random algorithm only for the greatest
           k=rank eigenvalues.
    :param string smooth_method: (default value = "none") method of smoothing

           - 'trigram' the 3-Gram trigram dict
             is computed and used by the predict function,
             in this case the threeGram probability is used instead of Spectral
             probability in negative case

           - 'none' or  something else no smooth method is used
             in predict function.
    :param boolean mode_quiet: (default value = False) True for no
           output message.

    :Example:

    >>> from splearn.spectral import Spectral
    >>> sp = Spectral()
    >>> sp.set_params(partial=True, lcolumns=6, lrows=6, smooth_method='trigram')
    Spectral(lcolumns=6, lrows=6, mode_quiet=False, partial=True, rank=5,
     smooth_method='trigram', sparse=True, version='classic')
    >>> sp.fit(data.data)
    Start Hankel matrix computation
    End of Hankel matrix computation
    Start Building Automaton from Hankel matrix
    End of Automaton computation
    Spectral(lcolumns=6, lrows=6, partial=True, rank=5, smooth_method='trigram', sparse=True, version='classic')
    >>> sp.automaton.initial
    array([-0.00049249,  0.00304676, -0.04405996, -0.10765322, -0.08660063])
    >>> sp.predict(data.data)
    array([  4.38961058e-04,   1.10616861e-01,   1.35569353e-03, ...,
        4.66041996e-06,   4.68177275e-02,   5.24287604e-20])
    >>> sp.loss(data.data, normalize=True)
    -10.530029936056017
    >>> sp.score(data.data)
    10.530029936056017

    """
    def __init__(self,  rank=5, lrows=7, lcolumns=7,
                 version='classic', partial=True,
                 sparse=True, full_svd_calculation=False, smooth_method='none',
                 mode_quiet=False):
        self.version = version
        self.partial = partial
        self.sparse = sparse
        self.full_svd_calculation = full_svd_calculation
        self.lrows = lrows
        self.lcolumns = lcolumns
        self.rank = rank
        self.trigram = {}
        self.smooth_method = smooth_method
        self._rule_smooth_method(smooth_method)
        self.mode_quiet = mode_quiet
        self._automaton = None
        self._hankel = None

    def get_params(self, deep=True):
        # suppose this estimator has parameters
        """
        return parameters values of Spectral estimator

        - Output:

        :returns: parameters dictionary of Spectral estimator name : value
        :rtype: dict

        """
        return {"rank": self.rank, "version": self.version,
                 "lrows": self.lrows, "lcolumns": self.lcolumns,
                 "partial": self.partial,
                 "sparse": self.sparse,
                 "full_svd_calculation" : self.full_svd_calculation,
                 "smooth_method" : self.smooth_method,
                 "mode_quiet" : self.mode_quiet
                }
    @property
    def automaton(self):
        """Automaton build by the fit method. None by default"""
        return self._automaton

    @automaton.setter
    def automaton(self, automaton):
        pass
    
    @property
    def hankel(self):
        """Hankel build by the fit method. None by default"""
        return self._hankel

    @hankel.setter
    def hankel(self, hankel):
        pass

    def _rule_smooth_method(self, value):
        if self.smooth_method not in ['none', 'trigram']:
            warnings.warn("smooth method should be in ['none', 'trigram']",
                          UserWarning)
            self.smooth_method = 'none'
        if value == 'trigram':
            self.smooth = 1
        else:
            self.smooth = 0

    def set_params(self, **parameters):
        """set the values of  Spectral estimator parameters

        - Output:

        :returns: Spectral estimator with new parameters
        :rtype: Spectral
        """
        for parameter, value in parameters.items():
            self.__setattr__(parameter, value)
            if parameter == "smooth_method":
                self._rule_smooth_method(value)
        return self

    def fit(self, X, y=None):
        """Fit the model

        - Input:

        :param SplearnArray X: object of shape [n_samples,n_features]
               Training data
        :param ndarray y: (default value = None) not used by Spectral estimator
               numpy array of shape [n_samples] Target values


        - Output:

        :returns: Spectral itself with an automaton attribute instanced
                  returns an instance of self.
        :rtype: Spectral

        """
        check_array(X)

        if not isinstance(X, SplearnArray):
            self._hankel = None
            self._automaton = None
            return self
        X = self.polulate_dictionnaries(X)
        self._hankel = Hankel(sample_instance=X,
                         lrows=self.lrows, lcolumns=self.lcolumns,
                         version=self.version,
                         partial=self.partial, sparse=self.sparse,
                         full_svd_calculation = self.full_svd_calculation,
                         mode_quiet=self.mode_quiet)
        self._automaton = self._hankel.to_automaton(self.rank, self.mode_quiet)
        # for smooth option compute trigram dictionnary
        if self.smooth:
            self.trigram = self._threegramdict(X.sample)

        return self

    def polulate_dictionnaries(self, X):
        """Populates the *sample*, *pref*, *suff*, *fact* dictionnaries of X
                
        - Input:

        :param SplearnArray X: object of shape [n_samples,n_features]
               Training data
        
        """
        if not isinstance(X, SplearnArray):
            return X
        dsample = {}  # dictionary (word,count)
        dpref = {}  # dictionary (prefix,count)
        dsuff = {}  # dictionary (suffix,count)
        dfact = {}  # dictionary (factor,count)
        if self.partial:
            if isinstance(self.lrows, int):
                lrowsmax = self.lrows
                version_rows_int = True
            else:
                version_rows_int = False
                lrowsmax = self.lrows.__len__()
            if isinstance(self.lcolumns, int):
                lcolumnsmax = self.lcolumns
                version_columns_int = True
            else:
                lcolumnsmax = self.lcolumns.__len__()
                version_columns_int = False
            lmax = lrowsmax + lcolumnsmax

        for line in range(X.shape[0]):
            w = X[line, :]
            w = w[w >= 0]
            w = tuple([int(x) for x in w[0:]])
            dsample[w] = dsample[w] + 1 if w in dsample else 1
            if self.version == "prefix" or self.version == "classic":
                # empty word treatment for prefixe, suffix, and factor dictionnaries
                dpref[()] = dpref[()] + 1 if () in dpref else 1
            if self.version == "suffix" or self.version == "classic":
                dsuff[()] = dsuff[()] + 1 if () in dsuff else 1
            if self.version == "factor" or self.version == "suffix" \
                    or self.version == "prefix":
                dfact[()] = dfact[()] + len(w) + 1 if () in dfact else len(
                    w) + 1
            if self.partial:
                for i in range(len(w)):
                    if self.version == "classic":
                        if (version_rows_int is True and
                                        i + 1 <= lrowsmax) or \
                        (version_rows_int is False and
                                    w[:i + 1] in self.lrows):
                            dpref[w[:i + 1]] = \
                                dpref[w[:i + 1]] + 1 if w[
                                                        :i + 1] in dpref else 1
                        if (version_columns_int is True and
                                        i + 1 <= lcolumnsmax) or \
                                (version_columns_int is False and
                                         w[-( i + 1):] in self.lcolumns):
                            dsuff[w[-(i + 1):]] = dsuff[w[-(i + 1):]] + 1 \
                                if w[-(i + 1):] in dsuff else 1
                    if self.version == "prefix":
                        # dictionaries dpref is populated until
                        # lmax = lrows + lcolumns
                        # dictionaries dfact is populated until lcolumns
                        if ((version_rows_int is True or
                                     version_columns_int is True) and
                                        i + 1 <= lmax) or \
                                (version_rows_int is False and
                                     (w[:i + 1] in self.lrows)) or \
                                (version_columns_int is False and
                                     (w[:i + 1] in self.lcolumns)):
                            dpref[w[:i + 1]] = dpref[w[:i + 1]] + 1 \
                                if w[:i + 1] in dpref else 1
                        for j in range(i + 1, len(w) + 1):
                            if (version_columns_int is True and (
                                j - i) <= lmax) or \
                                    (version_columns_int is False and
                                         (w[i:j] in self.lcolumns)):
                                dfact[w[i:j]] = dfact[w[i:j]] + 1 \
                                    if w[i:j] in dfact else 1
                    if self.version == "suffix":
                        if ((version_rows_int is True or
                                     version_columns_int is True) and
                                    i <= lmax) or \
                                (version_rows_int is False and
                                     (w[-(i + 1):] in self.lrows)) or \
                                (version_columns_int is False and
                                     (w[-(i + 1):] in self.lcolumns)):
                            dsuff[w[-(i + 1):]] = dsuff[w[-(i + 1):]] + 1 \
                                if w[-(i + 1):] in dsuff else 1
                        for j in range(i + 1, len(w) + 1):
                            if (version_rows_int is True and (
                                j - i) <= lmax) or \
                                    (version_rows_int is False and
                                         (w[i:j] in self.lrows)):
                                dfact[w[i:j]] = dfact[w[i:j]] + 1 \
                                    if w[i:j] in dfact else 1
                    if self.version == "factor":
                        for j in range(i + 1, len(w) + 1):
                            if ((version_rows_int is True or
                                         version_columns_int is True) and
                                        (j - i) <= lmax) or \
                                    (version_rows_int is False and
                                         (w[i:j] in self.lrows)) or \
                                    (version_columns_int is False and
                                         (w[i:j] in self.lcolumns)):
                                dfact[w[i:j]] = \
                                    dfact[w[i:j]] + 1 if w[i:j] in dfact else 1

            else:  # not partial
                for i in range(len(w)):
                    dpref[w[:i + 1]] = dpref[w[:i + 1]] + 1 \
                        if w[:i + 1] in dpref else 1
                    dsuff[w[i:]] = dsuff[w[i:]] + 1 if w[i:] in dsuff else 1
                    for j in range(i + 1, len(w) + 1):
                        dfact[w[i:j]] = dfact[w[i:j]] + 1 \
                            if w[i:j] in dfact else 1
        X.sample = dsample
        if self.version == "classic":
            X.pref = dpref
            X.suff = dsuff
            X.fact = {}
        if self.version == "suffix":
            X.suff = dsuff
            X.fact = dfact
            X.pref = {}
        if self.version == "prefix":
            X.pref = dpref
            X.fact = dfact
            X.suff = {}
        if self.version == "factor":
            X.fact = dfact
            X.suff = {}
            X.pref = {}
        return X

    def _populate_sample_dict(self, X):
        dsample = {}  # dictionary (word,count)
        for line in range(X.shape[0]):
            w = X[line, :]
            w = w[w >= 0]
            w = tuple([int(x) for x in w[0:]])
            dsample[w] = dsample[w] + 1 if w in dsample else 1
        return dsample

    @property
    def trigram(self):
        """The trigram dictionary"""
        return self._trigram

    @trigram.setter
    def trigram(self, DPdict_values):
        if (not isinstance(DPdict_values, dict)):
            mess = "DPdict should be a dicionary.\n"
            mess += "Actual : " + str(DPdict_values)
            raise TypeError(mess)
        self._trigram = DPdict_values

    def _trigramprobability(self, sequence, trigram_test):
        prob = np.float64(1.0)
        seq = list(sequence)
        ngramseq = [-1, -1] + seq + [-2]
        if len(seq) < 0:
            return 0
        for start in range(len(ngramseq) - 2):
            end = start + 2
            if tuple(ngramseq[start:end]) in self.trigram.keys():
                if ngramseq[end] in self.trigram[tuple(ngramseq[start:end])].keys():
                    val1_train = np.float64(
                        self.trigram[tuple(ngramseq[start:end])][ngramseq[end]])
                else:
                    val1_train = 0
                val2_train = np.float64(self.trigram[tuple(ngramseq[start:end])][-1])
            else:
                val1_train = -1
                val2_train = -1
            if tuple(ngramseq[start:end]) in trigram_test.keys():
                if ngramseq[end] in trigram_test[
                    tuple(ngramseq[start:end])].keys():
                    val1_test = np.float64(
                        trigram_test[tuple(ngramseq[start:end])][
                            ngramseq[end]])
                else:
                    val1_test = 0
                val2_test = np.float64(
                    trigram_test[tuple(ngramseq[start:end])][-1])
            else:
                val1_test = -1
                val2_test = -1
            if val1_test == -1 and val1_train == -1:
                return 0
            if val1_test == -1:
                prob = prob * val1_train / val2_train
            if val1_train == -1:
                prob = prob * val1_test / val2_test
            if val1_test != -1 and val1_train != -1:
                prob = prob * (val1_test + val1_train) / ( val2_test + val2_train)
        return prob

    def nb_trigram(self):
        """return the number of index affected by the trigram computation


        - Output:

        :returns: int number of trigram_index

        """

        try:
            nb = np.where(self.trigram_index == True)[0].shape[0]
            return nb
        except:
            warnings.warn(UserWarning, "trigram_index does not exist")
            pass


    @staticmethod
    def _threegramdict(sample):
        DPdict = dict()
        for sequence in sample.keys():
            seq = list(sequence)
            ngramseq = [-1, -1] + seq + [-2]
            for start in range(len(ngramseq) - 2):
                end = start + 2
                if tuple(ngramseq[start:end]) in DPdict:
                    table = DPdict[tuple(ngramseq[start:end])]
                    if ngramseq[end] in table:
                        table[ngramseq[end]] = table[ngramseq[end]] + sample[
                        sequence]
                    else:
                        table[ngramseq[end]] = sample[sequence]
                    table[-1] = table[-1] + sample[sequence]
                else:
                    table = dict()
                    table[ngramseq[end]] = sample[sequence]
                    table[-1] = sample[sequence]
                    DPdict[tuple(ngramseq[start:end])] = table
        return DPdict

    def predict(self, X):
        """Predict using the Spectral model

        - Input:

        :param SplearnArray X : of shape data shape = (n_samples, n_features)
               Samples.


        - Output:

        :returns: Probability corresponding to the input X,
                  array-like of shape = n_samples
        :rtype: ndarray
        """

        check_array(X)
        if not hasattr(self, 'automaton'):
            raise NotFittedError("This %(name)s instance is not fitted "
                                 "yet" % {'name': type(self).__name__})
        if self._automaton is None:
            print("No Automaton has been computed, "
                  "check the format of the input fit data")
            warnings.warn("check the format of the input fit data", UserWarning)
            return X

        Y = self.predict_proba(X)
        return Y

    def predict_proba(self, X):
        """
        Predict probability using the Spectral model

        - Input:

        :param SplearnArray X : Samples, data shape = (n_samples, n_features)


        - Output:

        :returns: Probability corresponding to the input X
                  of shape = (n_samples)
        :rtype: ndarray
        """
        #check_is_fitted(self, "classes_")
        X = check_array(X)
        if not hasattr(self, 'automaton'):
            raise NotFittedError("This %(name)s instance is not fitted "
                                 "yet" % {'name': type(self).__name__})

        # if Automaton is None because the fit pass through doing nothing
        if self._automaton is None:
            print("No Automaton has been computed, "
                  "check the format of the input fit data")
            warnings.warn("check the format of the input fit data", UserWarning)
            return X[:,0]
        # if self.smooth == 1 and self.trigram == {}:
        #     warnings.warn("Incompatibility of smooth_method "
        #                   " activate trigram smooth option in predictor "
        #                   " and fit again", UserWarning)
        #     self.trigram = self._threegramdict(X.sample)
        if self.smooth == 1:
            test_sample = self._populate_sample_dict(X=X)
            trigram_test = self._threegramdict(test_sample)
            trigram_index = np.zeros(X.shape[0], dtype=bool)
        Y = np.zeros(X.shape[0])
        i = 0
        for line in range(X.shape[0]):
            w = X[line, :]
            w = w[w >= 0]
            w = tuple([int(x) for x in w[0:]])
            val = self._automaton.val(w)
            if self.smooth == 1 and val <= 0:
                Y[i] = self._trigramprobability(w, trigram_test)
                trigram_index[i] = True
            else:
                Y[i] = val
            i += 1
        if self.smooth == 1:
            self.trigram_index = trigram_index
        return Y

    def loss(self, X, y=None, normalize=True):
        """Log probability using the Spectral model

        - Input:

        :param SplearnArray X: of shape data shape = (n_samples, n_features)
               Samples. X is validation data.
        :param ndarray y: (default value = Null)
               numpy array of shape [n_samples] Target values,
               is the ground truth target for X (in the supervised case) or
               None (in the unsupervised case)
        :param boolean normalize: (default value = True) calculation are
               performed and normalize by the number of sample in case of True

        - Output:

        :returns: mean of Log Probability corresponding to the input X
        :rtype: float
        """
        warnings.simplefilter("error", RuntimeWarning)
        predict_prob = self.predict_proba(X)
        if y is None:
            try:
                if normalize:
                    Y = np.mean(-np.log(predict_prob))
                else:
                    Y = np.sum(-np.log(predict_prob))
            except:
                msg = "function loss or score use log " + \
                      "function, values can't be" + \
                      " negative, use it with smooth_method" + \
                      " to avoid such problem"
                raise ValueError(msg)
            return Y
        else:
            if normalize:
                Y = np.mean((np.subtract(predict_prob, y) ** 2.0))
            else:
                Y = np.sum((np.subtract(predict_prob, y) ** 2.0))
            return Y


    def score(self, X, y=None, scoring="perplexity"):
        """score of the input target

        - Input:

        :param SplearnArray X: of shape data shape = (n_samples, n_features)
               Samples.
        :param ndarray y: (default value = None)
               numpy array of shape [n_samples] Target values,
               is the ground truth target for X (in the supervised case) or
               None (in the unsupervised case)
        :param string scoring: (default value = "perplexity")
               method for score computation

        - Output:

        :returns: score, on the input X
        :rtype: float
        """


        if scoring == "perplexity":
            if y is None:
                return - self.loss(X, y, normalize=True)
            else:
                predict_prob = self.predict_proba(X)
                sA, sC = 0, 0
                sA = sum(predict_prob)
                sC = sum(y)
                s = 0
                perplexity = 0
                for i in range(X.shape[0]):
                    try:
                        s = s + y[i] / sC * math.log(predict_prob[i] / sA)
                        perplexity = math.exp(-s)
                    except:
                        msg = "function loss or score use log " + \
                              "function values can't be" + \
                              " negative, use it with smooth_method" + \
                              "to avoid such problem"
                        raise ValueError(msg)
                return perplexity
        else:
            return - self.loss(X, y, normalize=True)


# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2016-2018
# -----------------
#
# * LabEx Archimède: http://labex-archimede.univ-amu.fr/
# * Laboratoire d'Informatique et Systèmes : http://www.lis-lab.fr/
#
# Contributors:
# ------------
#
# * François Denis <francois.denis_AT_lis-lab.fr>
# * Rémi Eyraud <remi.eyraud_AT_lis-lab.fr>
# * Denis Arrivault <contact.dev_AT_lis-lab.fr>
# * Dominique Benielli <dominique.benielli_AT_univ-amu.fr>
#
# Description:
# -----------
#
# scitkit-splearn is a toolbox in
# python for spectral learning algorithms.
#
# Version:
# -------
#
# * splearn version = 1.2.1
#
# Licence:
# -------
#
# License: 3-clause BSD
#
#
# ######### COPYRIGHT #########

from __future__ import division, print_function
import unittest
import numpy as np
import scipy.sparse as sps


from splearn.datasets.base import load_data_sample
from splearn.automaton import Automaton
from splearn.spectral import Spectral
from splearn.hankel import Hankel
from splearn.tests.datasets.get_dataset_path import get_dataset_path

class HankelTest(unittest.TestCase):

    def test_CreateHankel_classic(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False)
        cl.polulate_dictionnaries(data.data)
        lprefix = [()]
        lprefix = lprefix + [(i,) for i in range(data.data.nbL)]
        lprefix = lprefix+[(i, j) for i in range(data.data.nbL)
                           for j in range(data.data.nbL)]
        lsuffix = [()]
        lsuffix = lsuffix + [(i,) for i in range(data.data.nbL)]
        lsuffix = lsuffix + [(i, j) for i in range(data.data.nbL)
                             for j in range(data.data.nbL)]

        for s in data.data.pref.keys():
            self.assertTrue( s in lprefix  )
        for s in data.data.suff.keys():
            self.assertTrue( s in lsuffix )
        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
            version="classic", partial=False, sparse=False).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = np.array([[ 1.,  1.,  1.,  2.,  1.],
              [ 1.,  0.,  2.,  0.,  0.],
              [ 1.,  0.,  1.,  0.,  0.],
              [ 2.,  0.,  0.,  0.,  0.],
              [ 1.,  0.,  0.,  0.,  0.]])

        h_1 = np.array([[ 1.,  0.,  2.,  0.,  0.],
              [ 0.,  0.,  0.,  0.,  0.],
              [ 0.,  0.,  0.,  0.,  0.],
              [ 0.,  0.,  0.,  0.,  0.],
              [ 0.,  0.,  0.,  0.,  0.]])

        h_2 = np.array([[ 1.,  0.,  1.,  0.,  0.],
              [ 2.,  0.,  0.,  0.,  0.],
              [ 1.,  0.,  0.,  0.,  0.],
              [ 0.,  0.,  0.,  0.,  0.],
              [ 0.,  0.,  0.,  0.,  0.]])

        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)

        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="classic", partial=False, sparse=True).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = sps.dok_matrix(np.array([[1., 1., 1., 2., 1.],
                        [1., 0., 2., 0., 0.],
                        [1., 0., 1., 0., 0.],
                        [2., 0., 0., 0., 0.],
                        [1., 0., 0., 0., 0.]]))

        h_1 = sps.dok_matrix(np.array([[1., 0., 2., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]]))

        h_2 = sps.dok_matrix(np.array([[1., 0., 1., 0., 0.],
                        [2., 0., 0., 0., 0.],
                        [1., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]]))

        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)


    def test_CreateHankel_prefix(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, version="prefix")
        cl.polulate_dictionnaries(data.data)
        lprefix = [()]
        lprefix = lprefix + [(i,) for i in range(data.data.nbL)]
        lprefix = lprefix + [(i, j) for i in range(data.data.nbL)
                             for j in range(data.data.nbL)]
        lsuffix = [()]
        lsuffix = lsuffix + [(i,) for i in range(data.data.nbL)]
        lsuffix = lsuffix + [(i, j) for i in range(data.data.nbL)
                             for j in range(data.data.nbL)]

        for s in data.data.pref.keys():
            self.assertTrue(s in lprefix)
        for s in data.data.suff.keys():
            self.assertTrue(s in lsuffix)
        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="prefix", partial=False,
                     sparse=False).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = np.array([[ 6.,  3.,  2.,  2.,  1.],
                        [ 3.,  0.,  2.,  0.,  0.],
                        [ 2.,  0.,  1.,  0.,  0.],
                        [ 2.,  0.,  0.,  0.,  0.],
                        [ 1.,  0.,  0.,  0.,  0.]])


        h_1 = np.array([[ 3.,  0.,  2.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.]])


        h_2 = np.array([[ 2.,  0.,  1.,  0.,  0.],
                        [ 2.,  0.,  0.,  0.,  0.],
                        [ 1.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.]])


        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)

        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="prefix", partial=False, sparse=True).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = sps.dok_matrix(np.array([[ 6.,  3.,  2.,  2.,  1.],
                        [ 3.,  0.,  2.,  0.,  0.],
                        [ 2.,  0.,  1.,  0.,  0.],
                        [ 2.,  0.,  0.,  0.,  0.],
                        [ 1.,  0.,  0.,  0.,  0.]]))

        h_1 = sps.dok_matrix(np.array([[ 3.,  0.,  2.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.]]))

        h_2 = sps.dok_matrix(np.array([[ 2.,  0.,  1.,  0.,  0.],
                        [ 2.,  0.,  0.,  0.,  0.],
                        [ 1.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.],
                        [ 0.,  0.,  0.,  0.,  0.]]))

        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)

    def test_CreateHankel_suffix(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, version="suffix")
        cl.polulate_dictionnaries(data.data)
        lprefix = [()]
        lprefix = lprefix + [(i,) for i in range(data.data.nbL)]
        lprefix = lprefix + [(i, j) for i in range(data.data.nbL)
                             for j in range(data.data.nbL)]
        lsuffix = [()]
        lsuffix = lsuffix + [(i,) for i in range(data.data.nbL)]
        lsuffix = lsuffix + [(i, j) for i in range(data.data.nbL)
                             for j in range(data.data.nbL)]

        for s in data.data.pref.keys():
            self.assertTrue(s in lprefix)
        for s in data.data.suff.keys():
            self.assertTrue(s in lsuffix)
        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="suffix", partial=False,
                     sparse=False).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = np.array([[6., 1., 4., 2., 1.],
                        [1., 0., 2., 0., 0.],
                        [4., 0., 1., 0., 0.],
                        [2., 0., 0., 0., 0.],
                        [1., 0., 0., 0., 0.]])

        h_1 = np.array([[1., 0., 2., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]])

        h_2 = np.array([[4., 0., 1., 0., 0.],
                        [2., 0., 0., 0., 0.],
                        [1., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]])

        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)

        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="suffix", partial=False, sparse=True).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = sps.dok_matrix(np.array([[6., 1., 4., 2., 1.],
                                       [1., 0., 2., 0., 0.],
                                       [4., 0., 1., 0., 0.],
                                       [2., 0., 0., 0., 0.],
                                       [1., 0., 0., 0., 0.]]))

        h_1 = sps.dok_matrix(np.array([[1., 0., 2., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.]]))

        h_2 = sps.dok_matrix(np.array([[4., 0., 1., 0., 0.],
                                       [2., 0., 0., 0., 0.],
                                       [1., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.]]))

        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)

    def test_CreateHankel_factor(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral(partial=False, version="factor")
        cl.polulate_dictionnaries(data.data)
        lprefix = [()]
        lprefix = lprefix + [(i,) for i in range(data.data.nbL)]
        lprefix = lprefix + [(i, j) for i in range(data.data.nbL)
                             for j in range(data.data.nbL)]
        lsuffix = [()]
        lsuffix = lsuffix + [(i,) for i in range(data.data.nbL)]
        lsuffix = lsuffix + [(i, j) for i in range(data.data.nbL)
                             for j in range(data.data.nbL)]

        for s in data.data.pref.keys():
            self.assertTrue(s in lprefix)
        for s in data.data.suff.keys():
            self.assertTrue(s in lsuffix)
        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="factor", partial=False,
                     sparse=False).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = np.array([[14., 3., 5., 2., 1.],
                        [3., 0., 2., 0., 0.],
                        [5., 0., 1., 0., 0.],
                        [2., 0., 0., 0., 0.],
                        [1., 0., 0., 0., 0.]])

        h_1 = np.array([[3., 0., 2., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]])

        h_2 = np.array([[5., 0., 1., 0., 0.],
                        [2., 0., 0., 0., 0.],
                        [1., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0.]])

        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)

        l_h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="factor", partial=False, sparse=True).lhankel
        self.assertEqual(3, l_h.__len__())
        h_0 = sps.dok_matrix(np.array([[14., 3., 5., 2., 1.],
                                       [3., 0., 2., 0., 0.],
                                       [5., 0., 1., 0., 0.],
                                       [2., 0., 0., 0., 0.],
                                       [1., 0., 0., 0., 0.]]))

        h_1 = sps.dok_matrix(np.array([[3., 0., 2., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.]]))

        h_2 = sps.dok_matrix(np.array([[5., 0., 1., 0., 0.],
                                       [2., 0., 0., 0., 0.],
                                       [1., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.],
                                       [0., 0., 0., 0., 0.]]))

        np.testing.assert_equal(l_h[0], h_0)
        np.testing.assert_equal(l_h[1], h_1)
        np.testing.assert_equal(l_h[2], h_2)

    def test_Attr_Hankel(self):
        adr = get_dataset_path("essai")
        # adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        cl = Spectral()
        cl.polulate_dictionnaries(data.data)
        h = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
            version="classic", partial=False, sparse=False)
        with self.assertRaises(TypeError):
            h.nbL = 2.0
        # nbEx assignment
        with self.assertRaises(TypeError):
            h.nbEx = 6.8

    def testEqualityOperator(self):
        adr = get_dataset_path("essai")
        data = load_data_sample(adr=adr)
        cl = Spectral()
        cl.polulate_dictionnaries(data.data)
        h1 = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                    version="classic", partial=False, sparse=False)
        h2 = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                     version="prefix", partial=False, sparse=False)
        self.assertTrue(h1 != h2)
        h2 = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                    version="classic", partial=True, sparse=False)
        self.assertTrue(h1 != h2)
        h2 = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                    version="classic", partial=False, sparse=True)
        self.assertTrue(h1 != h2)
        h2 = Hankel(sample_instance=data.data, lrows=1, lcolumns=1,
                    version="classic", partial=False, sparse=False)
        h2.lhankel[0][0,0] -= 1.0
        self.assertTrue(h1 != h2)
        h2.lhankel = h2.lhankel[1:]
        self.assertTrue(h1 != h2)
        h2.nbEx -= 1
        self.assertTrue(h1 != h2)
        h2.nbL -= 1
        self.assertTrue(h1 != h2)
        l = [(0,), (1,), (0,0), (0,1), (1,0), (1,1)]
        h2 = h1.to_automaton(2).to_hankel(lrows=l, lcolumns=l)
        self.assertTrue(h1 != h2)
        
        adr = get_dataset_path("3.pautomac.train")
        data = load_data_sample(adr=adr)
        X = data.data
        sp1 = Spectral()
        sp1 = sp1.fit(X)
        H1 = sp1.hankel
        sp2 = Spectral()
        sp2 = sp2.fit(X)
        H2 = sp2.hankel
        H2.lhankel[0][(1, 1610)] = 0
        self.assertTrue(H1 != H2)
        

if __name__ == '__main__':
    unittest.main()
